"""Message storage.

The sequencing system we use in the message store may be quite confusing
if you haven't looked at it in the last 10 minutes. For that reason, let's
review the mechanics here.

Our goal is to implement a reasonably robust system for delivering messages
from us to our peer. The system should be smart enough to recover if the peer
happens to lose messages that we have already sent, provided that these
messages are not too old (we'll see below what 'too old' means).

Messages added to the store are identified by increasing natural numbers, the
first message added is identified by 0, the second by 1, and so on. We call
"sequence" the number identifying the next message that we want to send. For
example, if the store has been added ten messages (that we represent with
uppercase letters) and we want start sending the first of them, our store
would like like::

    sequence: 0
    messages: A, B, C, D, E, F, G, H, I, J
              ^

The "^" marker is what we call "pending offset" and is the displacement of the
message we want to send next from the first message we have in the store.

Let's say we now send to our peer a batch of 3 sequential messages. In the
payload we include the body of the messages being sent and the sequence, which
identifies the first message of the batch. In this case the payload would look
like (pseudo-code)::

    (sequence: 0, messages: A, B, C)

If everything works fine on the other end, our peer replies with a payload that
would like::

    (next-expected-sequence: 4)

meaning that the peer has received all the three messages that we sent, and so
the next message it expects to receive is the one identified by the number 4.
At this point we update both our pending offset and our sequence values, and
the store now looks like::

    sequence: 4
    messages: A, B, C, D, E, F, G, H, I, J
                       ^

Great, now let's pretend that we send another batch, this time with five
messages::

    (sequence: 4, messages: D, E, F, G, H)

Our peer receives them fine responding with a payload looking like::

    (next-expected-sequence: 9)

meaning that it received all the eight messages we sent so far and it's waiting
for the ninth. This is the second successful batch that we send in a row, so we
can be reasonably confident that at least the messages in the first batch are
not really needed anymore. We delete them and we update our sequence and
pending offset accordingly::

    sequence: 9
    messages: D, E, F, G, H, I, J
                             ^

Note that we still want to keep around the messages we sent in the very last
batch, just in case. Indeed we now try to send a third batch with the last two
messages that we have, but our peer surprisingly replies us with this payload::

    (next-expected-sequence: 6)

Ouch! This means that something bad happened and our peer has somehow lost not
only the two messages that we sent in the last batch, but also the last three
messages of the former batch :(

Luckly we've kept enough old messages around that we can try to send them
again, we update our sequence and pending offset and the store looks like::

    sequence: 6
    messages: D, E, F, G, H, I, J
                    ^

We can now start again sending messages using the same strategy.

Note however that in the worst case scenario we could receive from our peer
a next-expected-sequence value which is so old to be outside our buffer
of already-sent messages. In that case there is now way we can recover the
lost messages, and we'll just send the oldest one that we have.

See L{MessageStore} for details about how messages are stored on the file
system and L{landscape.lib.message.got_next_expected} to check how the
strategy for updating the pending offset and the sequence is implemented.
"""

import itertools
import logging
import os
import uuid

from landscape.lib import bpickle
from landscape.lib.fs import create_file
from landscape import SERVER_API


HELD = "h"
BROKEN = "b"


class MessageStore(object):
    """A message store which stores its messages in a file system hierarchy.

    Beside the "sequence" and the "pending offset" values described in the
    module docstring above, the L{MessageStore} also stores what we call
    "server sequence", which is the next message number expected by the
    *client* itself (because we are in turn the peer of a specular message
    system running in the server, which tries to deliver messages to us).

    The server sequence is entirely unrelated to the stored messages, but is
    incremented when successfully receiving messages from the server, in the
    very same way described above but with the roles inverted.

    @param persist: a L{Persist} used to save state parameters like the
        accepted message types, sequence, server uuid etc.
    @param directory: base of the file system hierarchy
    """

    api = SERVER_API

    def __init__(self, persist, directory, directory_size=1000):
        self._directory = directory
        self._directory_size = directory_size
        self._schemas = {}
        self._original_persist = persist
        self._persist = persist.root_at("message-store")
        message_dir = self._message_dir()
        if not os.path.isdir(message_dir):
            os.makedirs(message_dir)

    def commit(self):
        """Persist metadata to disk."""
        self._original_persist.save()

    def set_accepted_types(self, types):
        """Specify the types of messages that the server will expect from us.

        If messages are added to the store which are not currently
        accepted, they will be saved but ignored until their type is
        accepted.
        """
        assert type(types) in (tuple, list, set)
        self._persist.set("accepted-types", sorted(set(types)))
        self._reprocess_holding()

    def get_accepted_types(self):
        """Get a list of all accepted message types."""
        return self._persist.get("accepted-types", ())

    def accepts(self, type):
        """Return bool indicating if C{type} is an accepted message type."""
        return type in self.get_accepted_types()

    def get_sequence(self):
        """Get the current sequence.

        @return: The sequence number of the message that the server expects us
            to send on the next exchange.
        """
        return self._persist.get("sequence", 0)

    def set_sequence(self, number):
        """Set the current sequence.

        Set the sequence number of the message that the server expects us to
        send on the next exchange.
        """
        self._persist.set("sequence", number)

    def get_server_sequence(self):
        """Get the current server sequence.

        @return: the sequence number of the message that we will ask the server
            to send to us on the next exchange.
        """
        return self._persist.get("server_sequence", 0)

    def set_server_sequence(self, number):
        """Set the current server sequence.

        Set the sequence number of the message that we will ask the server to
        send to us on the next exchange.
        """
        self._persist.set("server_sequence", number)

    def get_server_uuid(self):
        """Return the currently set server UUID."""
        return self._persist.get("server_uuid")

    def set_server_uuid(self, uuid):
        """Change the known UUID from the server we're communicating to."""
        self._persist.set("server_uuid", uuid)

    def get_exchange_token(self):
        """Get the authentication token to use for the next exchange."""
        return self._persist.get("exchange_token")

    def set_exchange_token(self, token):
        """Set the authentication token to use for the next exchange."""
        self._persist.set("exchange_token", token)

    def get_pending_offset(self):
        """Get the current pending offset."""
        return self._persist.get("pending_offset", 0)

    def set_pending_offset(self, val):
        """Set the current pending offset.

        Set the offset into the message pool to consider assigned to the
        current sequence number as returned by l{get_sequence}.
        """
        self._persist.set("pending_offset", val)

    def add_pending_offset(self, val):
        """Increment the current pending offset by C{val}."""
        self.set_pending_offset(self.get_pending_offset() + val)

    def count_pending_messages(self):
        """Return the number of pending messages."""
        return sum(1 for x in self._walk_pending_messages())

    def get_pending_messages(self, max=None):
        """Get any pending messages that aren't being held, up to max."""
        accepted_types = self.get_accepted_types()
        messages = []
        for filename in self._walk_pending_messages():
            if max is not None and len(messages) >= max:
                break
            data = self._get_content(self._message_dir(filename))
            try:
                message = bpickle.loads(data)
            except ValueError, e:
                logging.exception(e)
                self._add_flags(filename, BROKEN)
            else:
                if message["type"] not in accepted_types:
                    self._add_flags(filename, HELD)
                else:
                    messages.append(message)
        return messages

    def delete_old_messages(self):
        """Delete messages which are unlikely to be needed in the future."""
        for fn in itertools.islice(self._walk_messages(exclude=HELD + BROKEN),
                                   self.get_pending_offset()):
            os.unlink(fn)
            containing_dir = os.path.split(fn)[0]
            if not os.listdir(containing_dir):
                os.rmdir(containing_dir)

    def delete_all_messages(self):
        """Remove ALL stored messages."""
        self.set_pending_offset(0)
        for filename in self._walk_messages():
            os.unlink(filename)

    def add_schema(self, schema):
        """Add a schema to be applied to messages of the given type.

        The schema must be an instance of L{landscape.schema.Message}.
        """
        self._schemas[schema.type] = schema

    def is_pending(self, message_id):
        """Return bool indicating if C{message_id} still hasn't been delivered.

        @param message_id: Identifier returned by the L{add()} method.
        """
        i = 0
        pending_offset = self.get_pending_offset()
        for filename in self._walk_messages(exclude=BROKEN):
            flags = self._get_flags(filename)
            if ((HELD in flags or i >= pending_offset) and
                os.stat(filename).st_ino == message_id):
                return True
            if BROKEN not in flags and HELD not in flags:
                i += 1
        return False

    def record_success(self, timestamp):
        """Record a successful exchange."""
        self._persist.remove("first-failure-time")
        self._persist.remove("blackhole-messages")

    def record_failure(self, timestamp):
        """
        Record a failed exchange, if all exchanges for the past week have
        failed then blackhole any future ones and request a full re-sync.
        """
        if not self._persist.has("first-failure-time"):
            self._persist.set("first-failure-time", timestamp)
        continued_failure_time = timestamp - self._persist.get(
            "first-failure-time")
        if self._persist.get("blackhole-messages"):
            # Already added the resync message
            return
        if continued_failure_time > (60 * 60 * 24 * 7):
            # reject all messages after a week of not exchanging
            self.add({"type": "resynchronize"})
            self._persist.set("blackhole-messages", True)
            logging.warning(
                "Unable to succesfully communicate with Landscape server "
                "for more than a week. Waiting for resync.")

    def add(self, message):
        """Queue a message for delivery.

        @param message: a C{dict} with a C{type} key and other keys conforming
            to the L{Message} schema for that specific message type.

        @return: message_id, which is an identifier for the added
                 message or C{None} if the message was rejected.
        """
        assert "type" in message
        if self._persist.get("blackhole-messages"):
            logging.debug("Dropped message, awaiting resync.")
            return
        message = self._schemas[message["type"]].coerce(message)

        if "api" not in message:
            message["api"] = self.api

        message_data = bpickle.dumps(message)

        filename = self._get_next_message_filename()
        temp_path = filename + ".tmp"
        create_file(temp_path, message_data)
        os.rename(temp_path, filename)

        if not self.accepts(message["type"]):
            filename = self._set_flags(filename, HELD)

        # For now we use the inode as the message id, as it will work
        # correctly even faced with holding/unholding.  It will break
        # if the store is copied over for some reason, but this shouldn't
        # present an issue given the current uses.  In the future we
        # should have a nice transactional storage (e.g. sqlite) which
        # will offer a more strong primary key.
        message_id = os.stat(filename).st_ino

        return message_id

    def _get_next_message_filename(self):
        message_dirs = self._get_sorted_filenames()
        if message_dirs:
            newest_dir = message_dirs[-1]
        else:
            os.makedirs(self._message_dir("0"))
            newest_dir = "0"

        message_filenames = self._get_sorted_filenames(newest_dir)
        if not message_filenames:
            filename = self._message_dir(newest_dir, "0")
        elif len(message_filenames) < self._directory_size:
            filename = str(int(message_filenames[-1].split("_")[0]) + 1)
            filename = self._message_dir(newest_dir, filename)
        else:
            newest_dir = self._message_dir(str(int(newest_dir) + 1))
            os.makedirs(newest_dir)
            filename = os.path.join(newest_dir, "0")

        return filename

    def _walk_pending_messages(self):
        """Walk the files which are definitely pending."""
        pending_offset = self.get_pending_offset()
        for i, filename in enumerate(
                self._walk_messages(exclude=HELD + BROKEN)):
            if i >= pending_offset:
                yield filename

    def _walk_messages(self, exclude=None):
        if exclude:
            exclude = set(exclude)
        message_dirs = self._get_sorted_filenames()
        for message_dir in message_dirs:
            for filename in self._get_sorted_filenames(message_dir):
                flags = set(self._get_flags(filename))
                if (not exclude or not exclude & flags):
                    yield self._message_dir(message_dir, filename)

    def _get_sorted_filenames(self, dir=""):
        message_files = [x for x in os.listdir(self._message_dir(dir))
                         if not x.endswith(".tmp")]
        message_files.sort(key=lambda x: int(x.split("_")[0]))
        return message_files

    def _message_dir(self, *args):
        return os.path.join(self._directory, *args)

    def _get_content(self, filename):
        file = open(filename)
        try:
            return file.read()
        finally:
            file.close()

    def _reprocess_holding(self):
        """
        Unhold accepted messages left behind, and hold unaccepted
        pending messages.
        """
        offset = 0
        pending_offset = self.get_pending_offset()
        accepted_types = self.get_accepted_types()
        for old_filename in self._walk_messages():
            flags = self._get_flags(old_filename)
            try:
                message = bpickle.loads(self._get_content(old_filename))
            except ValueError, e:
                logging.exception(e)
                if HELD not in flags:
                    offset += 1
            else:
                accepted = message["type"] in accepted_types
                if HELD in flags:
                    if accepted:
                        new_filename = self._get_next_message_filename()
                        os.rename(old_filename, new_filename)
                        self._set_flags(new_filename, set(flags) - set(HELD))
                else:
                    if not accepted and offset >= pending_offset:
                        self._set_flags(old_filename, set(flags) | set(HELD))
                    offset += 1

    def _get_flags(self, path):
        basename = os.path.basename(path)
        if "_" in basename:
            return basename.split("_")[1]
        return ""

    def _set_flags(self, path, flags):
        dirname, basename = os.path.split(path)
        new_path = os.path.join(dirname, basename.split("_")[0])
        if flags:
            new_path += "_" + "".join(sorted(set(flags)))
        os.rename(path, new_path)
        return new_path

    def _add_flags(self, path, flags):
        self._set_flags(path, self._get_flags(path) + flags)

    def get_session_id(self, scope=None):
        """Generate a unique session identifier, persist it and return it.

        See also L{landscape.broker.server.BrokerServer.get_session_id} for
        more information on what this is used for.

        @param scope: A string identifying the scope of interest of requesting
            object. Currently this is unused but it has been implemented in
            preparation for a fix for bug #300278 so that we don't have to
            change the persisted structure later.  When that fix is in place
            this will allow us to re-synchronise only certain types of
            information, limited by scope.
        """
        session_ids = self._persist.get("session-ids", {})
        for session_id, stored_scope in session_ids.iteritems():
            # This loop should be relatively short as it's intent is to limit
            # session-ids to one per scope.  The or condition here is not
            # strictly necessary, but we *should* do "is" comparisons when we
            # can (so says PEP 8).
            if scope is stored_scope or scope == stored_scope:
                return session_id
        session_id = str(uuid.uuid4())
        session_ids[session_id] = scope
        self._persist.set("session-ids", session_ids)
        return session_id

    def is_valid_session_id(self, session_id):
        """
        Returns L{True} if the provided L{session_id} is known by this
        L{MessageStore}.
        """
        return session_id in self._persist.get("session-ids", {})

    def drop_session_ids(self, scopes=None):
        """Drop all session ids."""
        new_session_ids = {}
        if scopes:
            session_ids = self._persist.get("session-ids", {})
            for session_id, session_scope in session_ids.iteritems():
                if session_scope not in scopes:
                    new_session_ids[session_id] = session_scope
        self._persist.set("session-ids", new_session_ids)


def get_default_message_store(*args, **kwargs):
    """
    Get a L{MessageStore} object with all Landscape message schemas added.
    """
    from landscape. message_schemas import message_schemas
    store = MessageStore(*args, **kwargs)
    for schema in message_schemas.values():
        store.add_schema(schema)
    return store



import logging

from gettext import gettext as _

from landscape.ui.constants import NOT_MANAGED, CANONICAL_MANAGED

from landscape.ui.model.registration.proxy import RegistrationProxy
from landscape.ui.model.configuration.state import StateError


class ConfigControllerLockError(Exception):
    pass


class ConfigController(object):
    """
    L{ConfigContoller} defines actions to take against a configuration object,
    providing starting values from the file, allowing them to be changed
    transiently, reverted or committed.
    """

    DEFAULT_DEDICATED_ACCOUNT_NAME = "standalone"

    def __init__(self, configuration):
        self._observers = []
        self._configuration = configuration
        self._initialised = True

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            return getattr(self._configuration, name)

    def __setattr__(self, name, value):
        # this test allows attributes to be set in the __init__ method
        if not '_initialised' in self.__dict__:
            return object.__setattr__(self, name, value)
        if name in ConfigController.__dict__:
            return object.__setattr__(self, name, value)
        else:
            try:
                setattr(self._configuration, name, value)
                self._configuration.modify()
            except AttributeError:
                return object.__setattr__(self, name, value)
            else:
                self._configuration.modify()

    def load(self):
        """
        Load the initial data from the configuration.
        """
        return self._configuration.load_data()

    def revert(self):
        """
        Revert settings to those the configuration object originally found.
        """
        try:
            self._configuration.revert()
        except StateError:
            # We probably don't care.
            logging.info("landscape-client-settings-ui reverted with no "
                         "changes to revert.")

    def persist(self, on_notify, on_error, on_succeed, on_fail):
        """Persist settings via the configuration object."""
        try:
            self._configuration.persist()
        except StateError:
            # We probably don't care.
            logging.info("landscape-client-settings-ui committed with no "
                         "changes to commit.")
        if self._configuration.management_type == NOT_MANAGED:
            self.disable(on_notify, on_succeed, on_fail)
        else:
            self.register(on_notify, on_error, on_succeed, on_fail)

    def register(self, notify_method, error_method, succeed_method,
                 fail_method):
        """
        Perform registration using the L{RegistrationProxy}.
        """

        def registration_fail_wrapper():
            fail_method(action=_("Registering client failed"))

        def registration_succeed_wrapper():
            succeed_method(action=_("Registering client was successful"))

        registration = RegistrationProxy(
            on_register_notify=notify_method,
            on_register_error=error_method,
            on_register_succeed=registration_succeed_wrapper,
            on_register_fail=registration_fail_wrapper)
        if self._configuration.management_type == CANONICAL_MANAGED:
            notify_method(_("Attempting to register at %s") %
                          self._configuration.hosted_landscape_host)
        else:
            notify_method(_("Attempting to register at %s") %
                          self._configuration.local_landscape_host)
        registration.register(self._configuration.get_config_filename())
        registration.exit()

    def disable(self, notify_method, succeed_method, fail_method):
        """
        Disable landscape client via the L{RegistrationProxy}.
        """

        def disabling_fail_wrapper():
            fail_method(action=_("Disabling client failed"))

        def disabling_succeed_wrapper():
            succeed_method(action=_("Disabling client was successful"))

        registration = RegistrationProxy(
            on_disable_succeed=disabling_succeed_wrapper,
            on_disable_fail=disabling_fail_wrapper)
        notify_method(_("Attempting to disable landscape client."))
        registration.disable()
        registration.exit()



import os

from landscape.broker.config import BrokerConfiguration
from landscape.tests.helpers import LandscapeTest, EnvironSaverHelper


class ConfigurationTests(LandscapeTest):

    helpers = [EnvironSaverHelper]

    def test_loading_sets_http_proxies(self):
        """
        The L{BrokerConfiguration.load} method sets the 'http_proxy' and
        'https_proxy' enviroment variables to the provided values.
        """
        if "http_proxy" in os.environ:
            del os.environ["http_proxy"]
        if "https_proxy" in os.environ:
            del os.environ["https_proxy"]

        configuration = BrokerConfiguration()
        configuration.load(["--http-proxy", "foo",
                            "--https-proxy", "bar",
                            "--url", "whatever"])
        self.assertEqual(os.environ["http_proxy"], "foo")
        self.assertEqual(os.environ["https_proxy"], "bar")

    def test_loading_without_http_proxies_does_not_touch_environment(self):
        """
        The L{BrokerConfiguration.load} method doesn't override the
        'http_proxy' and 'https_proxy' enviroment variables if they
        are already set and no new value was specified.
        """
        os.environ["http_proxy"] = "heyo"
        os.environ["https_proxy"] = "baroo"

        configuration = BrokerConfiguration()
        configuration.load(["--url", "whatever"])
        self.assertEqual(os.environ["http_proxy"], "heyo")
        self.assertEqual(os.environ["https_proxy"], "baroo")

    def test_loading_resets_http_proxies(self):
        """
        User scenario:

        Runs landscape-config, fat-fingers a random character into the
        http_proxy field when he didn't mean to. runs it again, this time
        leaving it blank. The proxy should be reset to whatever
        environment-supplied proxy there was at startup.
        """
        os.environ["http_proxy"] = "original"
        os.environ["https_proxy"] = "originals"

        configuration = BrokerConfiguration()
        configuration.load(["--http-proxy", "x",
                            "--https-proxy", "y",
                            "--url", "whatever"])
        self.assertEqual(os.environ["http_proxy"], "x")
        self.assertEqual(os.environ["https_proxy"], "y")

        configuration.load(["--url", "whatever"])
        self.assertEqual(os.environ["http_proxy"], "original")
        self.assertEqual(os.environ["https_proxy"], "originals")

    def test_default_exchange_intervals(self):
        """Exchange intervales are set to sane defaults."""
        configuration = BrokerConfiguration()
        self.assertEqual(60, configuration.urgent_exchange_interval)
        self.assertEqual(900, configuration.exchange_interval)

    def test_intervals_are_ints(self):
        """
        The 'urgent_exchange_interval, 'exchange_interval' and 'ping_interval'
        values specified in the configuration file are converted to integers.
        """
        filename = self.makeFile("[client]\n"
                                 "urgent_exchange_interval = 12\n"
                                 "exchange_interval = 34\n"
                                 "ping_interval = 6\n")

        configuration = BrokerConfiguration()
        configuration.load(["--config", filename, "--url", "whatever"])

        self.assertEqual(configuration.urgent_exchange_interval, 12)
        self.assertEqual(configuration.exchange_interval, 34)
        self.assertEqual(configuration.ping_interval, 6)

    def test_tag_handling(self):
        """
        The 'tags' value specified in the configuration file is not converted
        to a list (it must be a string). See bug #1228301.
        """
        filename = self.makeFile("[client]\n"
                                 "tags = check,linode,profile-test")

        configuration = BrokerConfiguration()
        configuration.load(["--config", filename, "--url", "whatever"])

        self.assertEqual(configuration.tags, "check,linode,profile-test")

    def test_access_group_handling(self):
        """
        The 'access_group' value specified in the configuration file is
        passed through.
        """
        filename = self.makeFile("[client]\n"
                                 "access_group = webserver")

        configuration = BrokerConfiguration()
        configuration.load(["--config", filename, "--url", "whatever"])

        self.assertEqual(configuration.access_group, "webserver")

    def test_missing_url_is_defaulted(self):
        """
        Test that if we don't explicitly pass a URL, then this value is
        defaulted.
        """
        filename = self.makeFile("[client]\n")

        configuration = BrokerConfiguration()
        configuration.load(["--config", filename])

        self.assertEqual(configuration.url,
                         "https://landscape.canonical.com/message-system")

    def test_server_autodiscover_handling(self):
        """
        server_autodiscover is parsed and converted to a boolean value by
        load().
        """
        configuration = BrokerConfiguration()
        configuration.load([])
        self.assertEqual(configuration.server_autodiscover, False)

        configuration = BrokerConfiguration()
        configuration.load(["--server-autodiscover=true"])
        self.assertEqual(configuration.server_autodiscover, True)

        configuration = BrokerConfiguration()
        configuration.load(["--server-autodiscover=false"])
        self.assertEqual(configuration.server_autodiscover, False)



"""Functions used by all sqlite-backed stores."""

try:
    import sqlite3
except ImportError:
    from pysqlite2 import dbapi2 as sqlite3


def with_cursor(method):
    """Decorator that encloses the method in a database transaction.

    Even though SQLite is supposed to be useful in autocommit mode, we've
    found cases where the database continued to be locked for writing
    until the cursor was closed.  With this in mind, instead of using
    the autocommit mode, we explicitly terminate transactions and enforce
    cursor closing with this decorator.
    """

    def inner(self, *args, **kwargs):
        if not self._db:
            # Create the database connection only when we start to actually
            # use it. This is essentially just a workaroud of a sqlite bug
            # happening when 2 concurrent processes try to create the tables
            # around the same time, the one which fails having an incorrect
            # cache and not seeing the tables
            self._db = sqlite3.connect(self._filename)
            self._ensure_schema()
        try:
            cursor = self._db.cursor()
            try:
                result = method(self, cursor, *args, **kwargs)
            finally:
                cursor.close()
            self._db.commit()
        except:
            self._db.rollback()
            raise
        return result
    return inner



from twisted.internet.defer import succeed

from landscape.lib.network import get_active_device_info


class Network(object):
    """Show information about active network interfaces.

    @param get_device_info: Optionally, a function that returns information
        about network interfaces.  Defaults to L{get_active_device_info}.
    """

    def __init__(self, get_device_info=None):
        if get_device_info is None:
            get_device_info = get_active_device_info
        self._get_device_info = get_device_info

    def register(self, sysinfo):
        """Register this plugin with the sysinfo system.

        @param sysinfo: The sysinfo registry.
        """
        self._sysinfo = sysinfo

    def run(self):
        """
        Gather information about network interfaces and write it to the
        sysinfo output.

        @return: A succeeded C{Deferred}.
        """
        for info in self._get_device_info():
            interface = info["interface"]
            ip_address = info["ip_address"]
            self._sysinfo.add_header("IP address for %s" % interface,
                                     ip_address)
        return succeed(None)



"""
Different versions of the Python DBus bindings return different types
to represent integers, strings, lists, etc.  Older versions return
builtin Python types: C{int}, C{str}, C{list}, etc.  Newer versions
return DBus-specific wrappers: C{Int16}, C{String}, C{Array}, etc.
Failures occur when DBus types are used because bpickle doesn't know
that an C{Int16} is really an C{int} and that an C{Array} is really a
C{list}.

L{install} and L{uninstall} can install and remove extensions that
make bpickle work with DBus types.
"""

import dbus

from landscape.lib import bpickle


def install():
    """Install bpickle extensions for DBus types."""
    for type, function in get_dbus_types():
        bpickle.dumps_table[type] = function


def uninstall():
    """Uninstall bpickle extensions for DBus types."""
    for type, function in get_dbus_types():
        del bpickle.dumps_table[type]


def dumps_utf8string(obj):
    """
    Convert the specified L{dbus.types.UTF8String} to bpickle's
    representation for C{unicode} data.
    """
    return "u%s:%s" % (len(obj), obj)


def dumps_double(obj):
    """
    Convert a dbus.types.Double into a floating point representation.
    """
    return "f%r;" % float(obj)


def get_dbus_types():
    """
    Generator yields C{(type, bpickle_function)} for available DBus
    types.
    """
    for (type_name, function) in [("Boolean", bpickle.dumps_bool),
                                  ("Int16", bpickle.dumps_int),
                                  ("UInt16", bpickle.dumps_int),
                                  ("Int32", bpickle.dumps_int),
                                  ("UInt32", bpickle.dumps_int),
                                  ("Int64", bpickle.dumps_int),
                                  ("UInt64", bpickle.dumps_int),
                                  ("Double", dumps_double),
                                  ("Array", bpickle.dumps_list),
                                  ("Dictionary", bpickle.dumps_dict),
                                  ("String", bpickle.dumps_unicode),
                                  ("UTF8String", dumps_utf8string)]:
        type = getattr(dbus.types, type_name, None)
        if type is not None:
            yield type, function



from landscape.ui.tests.helpers import (
    ConfigurationProxyHelper, dbus_test_should_skip, dbus_skip_message,
    FakeGSettings)

if not dbus_test_should_skip:
    from landscape.ui.controller.configuration import ConfigController
    import landscape.ui.model.configuration.state
    from landscape.ui.model.configuration.state import (
        ConfigurationModel, COMPUTER_TITLE)
    from landscape.ui.model.configuration.uisettings import UISettings

from landscape.tests.helpers import LandscapeTest


class ConfigControllerTest(LandscapeTest):

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = "\n".join(
            ["[client]",
             "data_path = /var/lib/landscape/client/",
             "http_proxy = http://proxy.localdomain:3192",
             "tags = a_tag",
             "url = https://landscape.canonical.com/message-system",
             "account_name = foo",
             "registration_key = bar",
             "computer_title = baz",
             "https_proxy = https://proxy.localdomain:6192",
             "ping_url = http://landscape.canonical.com/ping"])

        self.default_data = {"management-type": "canonical",
                             "computer-title": "",
                             "hosted-landscape-host": "",
                             "hosted-account-name": "",
                             "hosted-password": "",
                             "local-landscape-host": "",
                             "local-account-name": "",
                             "local-password": ""}

        super(ConfigControllerTest, self).setUp()
        landscape.ui.model.configuration.state.DEFAULT_DATA[COMPUTER_TITLE] \
            = "me.here.com"
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)

        def disable(on_suceed, on_fail):
            pass

        def register(on_notify, on_error, on_succeed, on_fail):
            pass

        self.controller = ConfigController(model)
        self.controller.disable = disable
        self.controller.register = register
        self.controller.load()

    def test_init(self):
        """
        Test that when we create a controller it has initial state read in
        directly from the configuration file.
        """
        self.controller.load()
        self.assertEqual("baz", self.controller.computer_title)
        self.assertTrue(self.controller.management_type)
        self.assertEqual("landscape.canonical.com",
                         self.controller.hosted_landscape_host)
        self.assertEqual("foo", self.controller.hosted_account_name)
        self.assertEqual("bar", self.controller.hosted_password)
        self.assertEqual("", self.controller.local_landscape_host)
        self.assertEqual("standalone", self.controller.local_account_name)
        self.assertEqual("", self.controller.local_password)

    def test_set_hosted_account_name(self):
        """
        Test that we can set the L{hosted_account_name} property.
        """
        self.controller.load()
        self.assertEqual(self.controller.hosted_account_name, "foo")
        self.controller.hosted_account_name = "shoe"
        self.assertEqual(self.controller.hosted_account_name, "shoe")

    def test_set_local_account_name(self):
        """
        Test that we can set the L{local_account_name} property.
        """
        self.controller.load()
        self.assertEqual("standalone", self.controller.local_account_name)
        self.controller.local_account_name = "shoe"
        self.assertEqual(self.controller.local_account_name, "shoe")

    def test_set_hosted_password(self):
        """
        Test that we can set the L{hosted_password} property.
        """
        self.controller.load()
        self.assertEqual(self.controller.hosted_password, "bar")
        self.controller.hosted_password = "nucker"
        self.assertEqual(self.controller.hosted_password, "nucker")

    def test_set_local_password(self):
        """
        Test that we can set the L{local_password} property.
        """
        self.controller.load()
        self.assertEqual(self.controller.local_password, "")
        self.controller.local_password = "nucker"
        self.assertEqual(self.controller.local_password, "nucker")

    def test_set_local_landscape_host(self):
        """
        Test that we can set the L{local_landscape_host} property.
        """
        self.controller.load()
        self.assertEqual("", self.controller.local_landscape_host)
        self.controller.local_landscape_host = "smelly.pants"
        self.assertEqual(self.controller.local_landscape_host, "smelly.pants")

    def test_revert(self):
        """
        Test that we can revert the controller to it's initial state.
        """
        self.controller.load()
        self.assertEqual(self.controller.hosted_account_name, "foo")
        self.controller.hosted_account_name = "Hildaborg"
        self.assertEqual(self.controller.hosted_account_name, "Hildaborg")
        self.controller.revert()
        self.assertEqual(self.controller.hosted_account_name, "foo")

    def test_is_modified(self):
        """
        Test that we can determine when something has been modified.
        """
        self.controller.load()
        self.assertFalse(self.controller.is_modified)
        self.controller.local_landscape_host = "bing.bang.a.bang"
        self.assertTrue(self.controller.is_modified)
        self.controller.revert()
        self.assertFalse(self.controller.is_modified)
        self.controller.account_name = "soldierBlue"
        self.assertTrue(self.controller.is_modified)
        self.controller.revert()
        self.assertFalse(self.controller.is_modified)
        self.controller.registration_key = "HesAnIndianCowboyInTheRodeo"
        self.assertTrue(self.controller.is_modified)

    def test_persist(self):
        """
        Test that we can write configuration settings back to the config file.
        """
        self.controller.load()
        self.assertEqual("", self.controller.local_landscape_host)
        self.controller.local_landscape_host = "landscape.localdomain"
        self.assertEqual("landscape.localdomain",
                         self.controller.local_landscape_host)
        self.controller.persist(None, None, None, None)
        self.assertEqual("landscape.localdomain",
                         self.controller.local_landscape_host)
        self.controller.local_landscape_host = "boo"
        self.controller.revert()
        self.assertEqual("landscape.localdomain",
                         self.controller.local_landscape_host)

    if dbus_test_should_skip:
        skip = dbus_skip_message


class EmptyConfigControllerTest(LandscapeTest):

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = ""
        self.default_data = {"management-type": "not",
                             "computer-title": "",
                             "hosted-landscape-host": "",
                             "hosted-account-name": "",
                             "hosted-password": "",
                             "local-landscape-host": "",
                             "local-account-name": "",
                             "local-password": ""}

        super(EmptyConfigControllerTest, self).setUp()
        landscape.ui.model.configuration.state.DEFAULT_DATA[COMPUTER_TITLE] \
            = "me.here.com"
        settings = FakeGSettings(data=self.default_data)
        uisettings = UISettings(settings)
        model = ConfigurationModel(proxy=self.proxy, uisettings=uisettings)
        self.controller = ConfigController(model)
        self.controller.load()

    def test_defaulting(self):
        """
        Test we set the correct values when loading a blank configuration.
        """
        self.controller.load()
        self.assertEqual("not", self.controller.management_type)
        self.assertEqual("", self.controller.hosted_account_name)
        self.assertEqual("", self.controller.hosted_password)
        self.assertEqual("landscape.canonical.com",
                         self.controller.hosted_landscape_host)
        self.assertEqual("standalone", self.controller.local_account_name)
        self.assertEqual("", self.controller.local_password)
        self.assertEqual("me.here.com", self.controller.computer_title)

    if dbus_test_should_skip:
        skip = dbus_skip_message



from cStringIO import StringIO
from ConfigParser import ConfigParser
import logging
import shutil
import pprint
import re
import os
import sys
import unittest


from logging import Handler, ERROR, Formatter
from twisted.trial.unittest import TestCase
from twisted.python.failure import Failure
from twisted.internet.defer import Deferred

from landscape.tests.subunit import run_isolated
from landscape.tests.mocker import MockerTestCase
from landscape.watchdog import bootstrap_list

from landscape.lib.persist import Persist

from landscape.reactor import FakeReactor

from landscape.deployment import BaseConfiguration
from landscape.broker.config import BrokerConfiguration
from landscape.broker.transport import FakeTransport
from landscape.monitor.config import MonitorConfiguration
from landscape.monitor.monitor import Monitor
from landscape.manager.manager import Manager

from landscape.broker.service import BrokerService
from landscape.broker.amp import FakeRemoteBroker, RemoteBrokerConnector
from landscape.manager.config import ManagerConfiguration


DEFAULT_ACCEPTED_TYPES = [
    "accepted-types", "registration", "resynchronize", "set-id",
    "set-intervals", "unknown-id"]


class HelperTestCase(unittest.TestCase):

    helpers = []

    def setUp(self):
        self._helper_instances = []
        if LogKeeperHelper not in self.helpers:
            self.helpers.insert(0, LogKeeperHelper)
        result = None
        for helper_factory in self.helpers:
            helper = helper_factory()
            if hasattr(helper, "set_up"):
                result = helper.set_up(self)
            self._helper_instances.append(helper)
        # Return the return value of the last helper, which
        # might be a deferred
        return result

    def tearDown(self):
        for helper in reversed(self._helper_instances):
            if hasattr(helper, "tear_down"):
                helper.tear_down(self)


class MessageTestCase(unittest.TestCase):

    def assertMessage(self, obtained, expected):
        obtained = obtained.copy()
        for key in ["api", "timestamp"]:
            if key not in expected and key in obtained:
                obtained.pop(key)
        if obtained != expected:
            raise self.failureException("Messages don't match.\n"
                                        "Expected:\n%s\nObtained:\n%s\n"
                                        % (pprint.pformat(expected),
                                           pprint.pformat(obtained)))

    def assertMessages(self, obtained, expected):
        self.assertEqual(type(obtained), list)
        self.assertEqual(type(expected), list)
        for obtained_message, expected_message in zip(obtained, expected):
            self.assertMessage(obtained_message, expected_message)
        obtained_len = len(obtained)
        expected_len = len(expected)
        diff = abs(expected_len - obtained_len)
        if obtained_len < expected_len:
            extra = pprint.pformat(expected[-diff:])
            raise self.failureException("Expected the following %d additional "
                                        "messages:\n%s" % (diff, extra))
        elif expected_len < obtained_len:
            extra = pprint.pformat(obtained[-diff:])
            raise self.failureException("Got %d more messages than expected:\n"
                                        "%s" % (diff, extra))


class LandscapeTest(MessageTestCase, MockerTestCase,
                    HelperTestCase, TestCase):

    def setUp(self):
        self._old_config_filenames = BaseConfiguration.default_config_filenames
        BaseConfiguration.default_config_filenames = [self.makeFile("")]
        MockerTestCase.setUp(self)
        TestCase.setUp(self)
        return HelperTestCase.setUp(self)

    def tearDown(self):
        BaseConfiguration.default_config_filenames = self._old_config_filenames
        TestCase.tearDown(self)
        HelperTestCase.tearDown(self)
        MockerTestCase.tearDown(self)

    def successResultOf(self, deferred):
        """See C{twisted.trial._synctest._Assertions.successResultOf}.

        This is a copy of the original method, which is available only
        since Twisted 12.3.0 (from 2012-12-20).
        """
        result = []
        deferred.addBoth(result.append)
        if not result:
            self.fail(
                "Success result expected on %r, found no result instead" % (
                    deferred,))
        elif isinstance(result[0], Failure):
            self.fail(
                "Success result expected on %r, "
                "found failure result (%r) instead" % (deferred, result[0]))
        else:
            return result[0]

    def failureResultOf(self, deferred):
        """See C{twisted.trial._synctest._Assertions.failureResultOf}.

        This is a copy of the original method, which is available only
        since Twisted 12.3.0 (from 2012-12-20).
        """
        result = []
        deferred.addBoth(result.append)
        if not result:
            self.fail(
                "Failure result expected on %r, found no result instead" % (
                    deferred,))
        elif not isinstance(result[0], Failure):
            self.fail(
                "Failure result expected on %r, "
                "found success result (%r) instead" % (deferred, result[0]))
        else:
            return result[0]

    def assertNoResult(self, deferred):
        """See C{twisted.trial._synctest._Assertions.assertNoResult}.

        This is a copy of the original method, which is available only
        since Twisted 12.3.0 (from 2012-12-20).
        """
        result = []
        deferred.addBoth(result.append)
        if result:
            self.fail(
                "No result expected on %r, found %r instead" % (
                    deferred, result[0]))

    def assertDeferredSucceeded(self, deferred):
        self.assertTrue(isinstance(deferred, Deferred))
        called = []

        def callback(result):
            called.append(True)
        deferred.addCallback(callback)
        self.assertTrue(called)

    def assertSuccess(self, deferred, result=None):
        """
        Assert that the given C{deferred} results in the given C{result}.
        """
        self.assertTrue(isinstance(deferred, Deferred))
        return deferred.addCallback(self.assertEqual, result)

    def assertFileContent(self, filename, expected_content):
        fd = open(filename)
        actual_content = fd.read()
        fd.close()
        self.assertEqual(expected_content, actual_content)

    def assertConfigEqual(self, first, second):
        """
        Compare two configuration files for equality.  The order of parameters
        and comments may be different but the actual parameters and sections
        must be the same.
        """
        first_fp = StringIO(first)
        first_parser = ConfigParser()
        first_parser.readfp(first_fp)

        second_fp = StringIO(second)
        second_parser = ConfigParser()
        second_parser.readfp(second_fp)

        self.assertEqual(set(first_parser.sections()),
                         set(second_parser.sections()))
        for section in first_parser.sections():
            self.assertEqual(dict(first_parser.items(section)),
                             dict(second_parser.items(section)))

    def makePersistFile(self, *args, **kwargs):
        """Return a temporary filename to be used by a L{Persist} object.

        The possible .old persist file is cleaned up after the test.

        @see: L{MockerTestCase.makeFile}
        """
        persist_filename = self.makeFile(*args, **kwargs)

        def remove_saved_persist():
            try:
                os.remove(persist_filename + ".old")
            except OSError:
                pass
        self.addCleanup(remove_saved_persist)
        return persist_filename


class LandscapeIsolatedTest(LandscapeTest):
    """TestCase that also runs all test methods in a subprocess."""

    def run(self, result):
        if not getattr(LandscapeTest, "_cleanup_patch", False):
            run_method = LandscapeTest.run

            def run_wrapper(oself, *args, **kwargs):
                try:
                    return run_method(oself, *args, **kwargs)
                finally:
                    MockerTestCase._MockerTestCase__cleanup(oself)
            LandscapeTest.run = run_wrapper
            LandscapeTest._cleanup_patch = True
        run_isolated(LandscapeTest, self, result)


class ErrorHandler(Handler):

    def __init__(self, *args, **kwargs):
        Handler.__init__(self, *args, **kwargs)
        self.errors = []

    def emit(self, record):
        if record.levelno >= ERROR:
            self.errors.append(record)


class LoggedErrorsError(Exception):

    def __str__(self):
        out = "The following errors were logged\n"
        formatter = Formatter()
        for error in self.args[0]:
            out += formatter.format(error) + "\n"
        return out


class LogKeeperHelper(object):
    """Record logging information.

    Puts a 'logfile' attribute on your test case, which is a StringIO
    containing all log output.
    """

    def set_up(self, test_case):
        self.ignored_exception_regexes = []
        self.ignored_exception_types = []
        self.error_handler = ErrorHandler()
        test_case.log_helper = self
        test_case.logger = logger = logging.getLogger()
        test_case.logfile = StringIO()
        handler = logging.StreamHandler(test_case.logfile)
        format = ("%(levelname)8s: %(message)s")
        handler.setFormatter(logging.Formatter(format))
        self.old_handlers = logger.handlers
        self.old_level = logger.level
        logger.handlers = [handler, self.error_handler]
        logger.setLevel(logging.NOTSET)

    def tear_down(self, test_case):
        logger = logging.getLogger()
        logger.setLevel(self.old_level)
        logger.handlers = self.old_handlers
        errors = []
        for record in self.error_handler.errors:
            for ignored_type in self.ignored_exception_types:
                if (record.exc_info and record.exc_info[0]
                    and issubclass(record.exc_info[0], ignored_type)):
                    break
            else:
                for ignored_regex in self.ignored_exception_regexes:
                    if ignored_regex.match(record.message):
                        break
                else:
                    errors.append(record)
        if errors:
            raise LoggedErrorsError(errors)

    def ignore_errors(self, type_or_regex):
        if isinstance(type_or_regex, basestring):
            self.ignored_exception_regexes.append(re.compile(type_or_regex))
        else:
            self.ignored_exception_types.append(type_or_regex)


class EnvironSnapshot(object):

    def __init__(self):
        self._snapshot = os.environ.copy()

    def restore(self):
        os.environ.update(self._snapshot)
        for key in list(os.environ):
            if key not in self._snapshot:
                del os.environ[key]


class EnvironSaverHelper(object):

    def set_up(self, test_case):
        self._snapshot = EnvironSnapshot()

    def tear_down(self, test_case):
        self._snapshot.restore()


class FakeBrokerServiceHelper(object):
    """
    The following attributes will be set in your test case:
      - broker_service: A C{BrokerService}.
      - remote: A C{FakeRemoteBroker} behaving like a L{RemoteBroker} connected
          to the broker serivice but performing all operation synchronously.
    """

    def set_up(self, test_case):
        test_case.data_path = test_case.makeDir()
        log_dir = test_case.makeDir()
        test_case.config_filename = test_case.makeFile(
            "[client]\n"
            "url = http://localhost:91919\n"
            "computer_title = Some Computer\n"
            "account_name = some_account\n"
            "ping_url = http://localhost:91910\n"
            "data_path = %s\n"
            "log_dir = %s\n" % (test_case.data_path, log_dir))

        bootstrap_list.bootstrap(data_path=test_case.data_path,
                                 log_dir=log_dir)

        config = BrokerConfiguration()
        config.load(["-c", test_case.config_filename])

        class FakeBrokerService(BrokerService):
            reactor_factory = FakeReactor
            transport_factory = FakeTransport

        test_case.broker_service = FakeBrokerService(config)
        test_case.remote = FakeRemoteBroker(
            test_case.broker_service.exchanger,
            test_case.broker_service.message_store,
            test_case.broker_service.broker)


class BrokerServiceHelper(FakeBrokerServiceHelper):
    """
    Provides what L{FakeBrokerServiceHelper} does, and makes it a
    'live' service using a real L{RemoteBroker} connected over AMP.

    This adds the following attributes to your test case:
     - remote: A connected L{RemoteBroker}.
    """

    def set_up(self, test_case):
        super(BrokerServiceHelper, self).set_up(test_case)
        test_case.broker_service.startService()
        # Use different reactor to simulate separate processes
        self._connector = RemoteBrokerConnector(
            FakeReactor(), test_case.broker_service.config)
        deferred = self._connector.connect()
        test_case.remote = test_case.successResultOf(deferred)

    def tear_down(self, test_case):
        self._connector.disconnect()
        test_case.broker_service.stopService()


class MonitorHelper(FakeBrokerServiceHelper):
    """
    Provides everything that L{FakeBrokerServiceHelper} does plus a
    L{Monitor} instance.
    """

    def set_up(self, test_case):
        super(MonitorHelper, self).set_up(test_case)
        persist = Persist()
        persist_filename = test_case.makePersistFile()
        test_case.config = MonitorConfiguration()
        test_case.config.load(["-c", test_case.config_filename])
        test_case.reactor = FakeReactor()
        test_case.monitor = Monitor(
            test_case.reactor, test_case.config,
            persist, persist_filename)
        test_case.monitor.broker = test_case.remote
        test_case.mstore = test_case.broker_service.message_store


class ManagerHelper(FakeBrokerServiceHelper):
    """
    Provides everything that L{FakeBrokerServiceHelper} does plus a
    L{Manager} instance.
    """

    def set_up(self, test_case):
        super(ManagerHelper, self).set_up(test_case)
        test_case.config = ManagerConfiguration()
        test_case.config.load(["-c", test_case.config_filename])
        test_case.reactor = FakeReactor()
        test_case.manager = Manager(test_case.reactor, test_case.config)
        test_case.manager.broker = test_case.remote


class MockPopen(object):

    def __init__(self, output, return_codes=None):
        self.output = output
        self.stdout = StringIO(output)
        self.popen_inputs = []
        self.return_codes = return_codes

    def __call__(self, args, stdout=None, stderr=None):
        return self.popen(args, stdout=stdout, stderr=stderr)

    def popen(self, args, stdout=None, stderr=None):
        self.popen_inputs.append(args)
        return self

    def wait(self):
        if self.return_codes is None:
            return 0
        return self.return_codes.pop(0)


class StandardIOHelper(object):

    def set_up(self, test_case):
        from StringIO import StringIO

        test_case.old_stdout = sys.stdout
        test_case.old_stdin = sys.stdin
        test_case.stdout = sys.stdout = StringIO()
        test_case.stdin = sys.stdin = StringIO()
        test_case.stdin.encoding = "UTF-8"

    def tear_down(self, test_case):
        sys.stdout = test_case.old_stdout
        sys.stdin = test_case.old_stdin


class MockCoverageMonitor(object):

    def __init__(self, count=None, expected_count=None, percent=None,
                 since_reset=None, warn=None):
        self.count = count or 0
        self.expected_count = expected_count or 0
        self.percent = percent or 0.0
        self.since_reset_value = since_reset or 0
        self.warn_value = bool(warn)

    def since_reset(self):
        return self.since_reset_value

    def warn(self):
        return self.warn_value

    def reset(self):
        pass


class MockFrequencyMonitor(object):

    def __init__(self, count=None, expected_count=None, warn=None):
        self.count = count or 0
        self.expected_count = expected_count or 0
        self.warn_value = bool(warn)

    def warn(self):
        return self.warn_value

    def reset(self):
        pass


def mock_counter(i=0):
    """Generator starts at zero and yields integers that grow by one."""
    while True:
        yield i
        i += 1


def mock_time():
    """Generator starts at 100 and yields int timestamps that grow by one."""
    return mock_counter(100)


class StubProcessFactory(object):
    """
    A L{IReactorProcess} provider which records L{spawnProcess} calls and
    allows tests to get at the protocol.
    """

    def __init__(self):
        self.spawns = []

    def spawnProcess(self, protocol, executable, args=(), env={}, path=None,
                    uid=None, gid=None, usePTY=0, childFDs=None):
        self.spawns.append((protocol, executable, args,
                            env, path, uid, gid, usePTY, childFDs))


class DummyProcess(object):
    """A process (transport) that doesn't do anything."""

    def __init__(self):
        self.signals = []

    def signalProcess(self, signal):
        self.signals.append(signal)

    def closeChildFD(self, fd):
        pass


class ProcessDataBuilder(object):
    """Builder creates sample data for the process info plugin to consume.

    @param sample_dir: The directory for sample data.
    """

    RUNNING = "R (running)"
    STOPPED = "T (stopped)"
    TRACING_STOP = "T (tracing stop)"
    DISK_SLEEP = "D (disk sleep)"
    SLEEPING = "S (sleeping)"
    DEAD = "X (dead)"
    ZOMBIE = "Z (zombie)"

    def __init__(self, sample_dir):
        self._sample_dir = sample_dir

    def create_data(self, process_id, state, uid, gid,
                    started_after_boot=0, process_name=None,
                    generate_cmd_line=True, stat_data=None, vmsize=11676):

        """Creates sample data for a process.

        @param started_after_boot: The amount of time, in jiffies,
            between the system uptime and start of the process.
        @param process_name: Used to generate the process name that appears in
            /proc/%(pid)s/status
        @param generate_cmd_line: If true, place the process_name in
            /proc/%(pid)s/cmdline, otherwise leave it empty (this simulates a
            kernel process)
        @param stat_data: Array of items to write to the /proc/<pid>/stat file.
        """
        sample_data = """
Name:   %(process_name)s
State:  %(state)s
Tgid:   24759
Pid:    24759
PPid:   17238
TracerPid:      0
Uid:    %(uid)d    0    0    0
Gid:    %(gid)d    0    0    0
FDSize: 256
Groups: 4 20 24 25 29 30 44 46 106 110 112 1000
VmPeak:    11680 kB
VmSize:    %(vmsize)d kB
VmLck:         0 kB
VmHWM:      6928 kB
VmRSS:      6924 kB
VmData:     1636 kB
VmStk:       196 kB
VmExe:      1332 kB
VmLib:      4240 kB
VmPTE:        20 kB
Threads:        1
SigQ:   0/4294967295
SigPnd: 0000000000000000
ShdPnd: 0000000000000000
SigBlk: 0000000000000000
SigIgn: 0000000000000000
SigCgt: 0000000059816eff
CapInh: 0000000000000000
CapPrm: 0000000000000000
CapEff: 0000000000000000
""" % ({"process_name": process_name[:15], "state": state, "uid": uid,
        "gid": gid, "vmsize": vmsize})
        process_dir = os.path.join(self._sample_dir, str(process_id))
        os.mkdir(process_dir)
        filename = os.path.join(process_dir, "status")

        file = open(filename, "w+")
        try:
            file.write(sample_data)
        finally:
            file.close()
        if stat_data is None:
            stat_data = """\
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 %d\
""" % (started_after_boot,)
        filename = os.path.join(process_dir, "stat")

        file = open(filename, "w+")
        try:
            file.write(stat_data)
        finally:
            file.close()

        if generate_cmd_line:
            sample_data = """\
/usr/sbin/%(process_name)s\0--pid-file\0/var/run/%(process_name)s.pid\0
""" % {"process_name": process_name}
        else:
            sample_data = ""
        filename = os.path.join(process_dir, "cmdline")

        file = open(filename, "w+")
        try:
            file.write(sample_data)
        finally:
            file.close()

    def remove_data(self, process_id):
        """Remove sample data for the process that matches C{process_id}."""
        process_dir = os.path.join(self._sample_dir, str(process_id))
        shutil.rmtree(process_dir)


class FakePersist(object):
    """
    Incompletely fake a C{landscape.lib.Persist} to simplify higher level tests
    that result in an attempt to clear down persisted data.
    """

    def __init__(self):
        self.called = False

    def remove(self, key):
        self.called = True



from landscape.configuration import LandscapeSetupConfiguration
from landscape.ui.tests.helpers import (
    ConfigurationProxyHelper, dbus_test_should_skip, dbus_skip_message)
if not dbus_test_should_skip:
    from landscape.ui.model.configuration.mechanism import (
        PermissionDeniedByPolicy)
    import dbus
from landscape.tests.helpers import LandscapeTest


class AuthenticationFailureTest(LandscapeTest):
    """
    Test that an authentication failure is handled correctly.
    """
    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = ""

        super(AuthenticationFailureTest, self).setUp()

    def test_failed_authentication(self):
        """
        Test that load returns False when authentication fails.
        """

        def fake_policy_failure_load(arglist):
            """
            This simulates what you see if you click "Cancel" or give the wrong
            credentials 3 times when L{PolicyKit} challenges you.
            """
            raise PermissionDeniedByPolicy()

        def fake_timeout_failure_load(arglist):
            """
            This simulates what you see if you take no action when L{PolicyKit}
            challenges you.
            """

            class FakeNoReply(dbus.DBusException):
                """
                Simulate a L{DBus} L{NoReply} exception.
                """
                _dbus_error_name = "org.freedesktop.DBus.Error.NoReply"

            raise FakeNoReply()

        self.mechanism.load = fake_policy_failure_load
        self.assertFalse(self.proxy.load([]))
        self.mechanism.load = fake_timeout_failure_load
        self.assertFalse(self.proxy.load([]))

    if dbus_test_should_skip:
        skip = dbus_skip_message


class ConfigurationProxyInterfaceTest(LandscapeTest):
    """
    Test that we define the correct interface to a
    L{LandscapeSetupConfiguration} by really using one as the interface.
    """

    helpers = [ConfigurationProxyHelper]

    def setUp(self):
        self.config_string = "[client]\n" \
            "data_path = /var/lib/landscape/client/\n" \
            "http_proxy = http://proxy.localdomain:3192\n" \
            "tags = a_tag\n" \
            "url = https://landscape.canonical.com/message-system\n" \
            "account_name = foo\n" \
            "registration_key = boink\n" \
            "computer_title = baz\n" \
            "https_proxy = https://proxy.localdomain:6192\n" \
            "ping_url = http://landscape.canonical.com/ping\n"

        super(ConfigurationProxyInterfaceTest, self).setUp()

    def test_method_docstrings(self):
        """
        Test that we pass through the docstrings for methods.
        """
        self.assertEqual(self.proxy.load.__doc__,
                         LandscapeSetupConfiguration.load.__doc__)
        self.assertEqual(self.proxy.reload.__doc__,
                         LandscapeSetupConfiguration.reload.__doc__)
        self.assertEqual(self.proxy.write.__doc__,
                         LandscapeSetupConfiguration.write.__doc__)

    def test_account_name(self):
        """
        Test that we can get and set an account name via the configuration
        proxy.
        """
        self.assertEqual("foo", self.proxy.account_name)
        self.proxy.account_name = "bar"
        self.assertEqual("bar", self.proxy.account_name)
        self.assertEqual("bar", self.config.account_name)

    def test_computer_title(self):
        """
        Test that we can get and set a computer title via the configuration
        proxy.
        """
        self.assertEqual("baz", self.proxy.computer_title)
        self.proxy.computer_title = "bar"
        self.assertEqual("bar", self.proxy.computer_title)
        self.assertEqual("bar", self.config.computer_title)

    def test_data_path(self):
        """
        Test that we can get and set the data path via the configuration proxy.
        """
        self.assertEqual("/var/lib/landscape/client/", self.proxy.data_path)
        self.proxy.data_path = "bar"
        self.assertEqual("bar", self.proxy.data_path)
        self.assertEqual("bar", self.config.data_path)

    def test_http_proxy(self):
        """
        Test that we can get and set the HTTP proxy via the configuration
        proxy.
        """
        self.assertEqual("http://proxy.localdomain:3192",
                         self.proxy.http_proxy)
        self.proxy.http_proxy = "bar"
        self.assertEqual("bar", self.proxy.http_proxy)
        self.assertEqual("bar", self.config.http_proxy)

    def test_https_proxy(self):
        """
        Test that we can get and set the HTTPS proxy via the configuration
        proxy.
        """
        self.assertEqual("https://proxy.localdomain:6192",
                         self.proxy.https_proxy)
        self.proxy.https_proxy = "bar"
        self.assertEqual("bar", self.proxy.https_proxy)
        self.assertEqual("bar", self.config.https_proxy)

    def test_ping_url(self):
        """
        Test that we can get and set the ping URL via the configuration proxy.
        """
        self.assertEqual("http://landscape.canonical.com/ping",
                         self.proxy.ping_url)
        self.proxy.ping_url = "bar"
        self.assertEqual("bar", self.proxy.ping_url)
        self.assertEqual("bar", self.config.ping_url)

    def test_registration_key(self):
        """
        Test that we can get and set the registration key via the
        configuration proxy.
        """
        self.assertEqual("boink", self.proxy.registration_key)
        self.proxy.registration_key = "bar"
        self.assertEqual("bar", self.proxy.registration_key)
        self.assertEqual("bar", self.config.registration_key)

    def test_tags(self):
        """
        Test that we can get and set the tags via the configuration proxy.
        """
        self.assertEqual("a_tag", self.proxy.tags)
        self.proxy.tags = "bar"
        self.assertEqual("bar", self.proxy.tags)
        self.assertEqual("bar", self.config.tags)

    def test_url(self):
        """
        Test that we can get and set the URL via the configuration proxy.
        """
        self.assertEqual("https://landscape.canonical.com/message-system",
                         self.proxy.url)
        self.proxy.url = "bar"
        self.assertEqual("bar", self.proxy.url)
        self.assertEqual("bar", self.config.url)

    def test_exit(self):
        """
        Test that we can cause the mechanism to exit.
        """
        self.assertRaises(SystemExit, self.proxy.exit, asynchronous=False)

    if dbus_test_should_skip:
        skip = dbus_skip_message



import os
import sys
import grp
import pwd
import shutil
import logging
import tarfile
import cStringIO
import ConfigParser

from twisted.internet.defer import succeed

from landscape.lib.fetch import url_to_filename, fetch_to_files
from landscape.lib.lsb_release import parse_lsb_release, LSB_RELEASE_FILENAME
from landscape.lib.gpg import gpg_verify
from landscape.lib.fs import read_file
from landscape.package.taskhandler import (
    PackageTaskHandlerConfiguration, PackageTaskHandler, run_task_handler)
from landscape.lib.twisted_util import spawn_process
from landscape.manager.manager import SUCCEEDED, FAILED
from landscape.package.reporter import find_reporter_command


class ReleaseUpgraderConfiguration(PackageTaskHandlerConfiguration):
    """Specialized configuration for the Landscape release-upgrader."""

    @property
    def upgrade_tool_directory(self):
        """
        The directory where the upgrade-tool files get stored and extracted.
        """
        return os.path.join(self.package_directory, "upgrade-tool")


class ReleaseUpgrader(PackageTaskHandler):
    """Perform release upgrades.

    @cvar config_factory: The configuration class to use to build configuration
        objects to be passed to our constructor.
    @cvar queue_name: The queue we pick tasks from.
    @cvar lsb_release_filename: The path to the LSB data on the file system.
    @cvar landscape_ppa_url: The URL of the Landscape PPA, if it is present
        in the computer's sources.list it won't be commented out.
    @cvar logs_directory: Path to the directory holding the upgrade-tool logs.
    @cvar logs_limit: When reporting upgrade-tool logs to the server, only the
        last C{logs_limit} lines will be sent.
    """

    config_factory = ReleaseUpgraderConfiguration
    queue_name = "release-upgrader"
    lsb_release_filename = LSB_RELEASE_FILENAME
    landscape_ppa_url = "http://ppa.launchpad.net/landscape/trunk/ubuntu/"
    logs_directory = "/var/log/dist-upgrade"
    logs_limit = 100000

    def make_operation_result_message(self, operation_id, status, text, code):
        """Convenience to create messages of type C{"operation-result"}."""
        return {"type": "operation-result",
                "operation-id": operation_id,
                "status": status,
                "result-text": text,
                "result-code": code}

    def handle_task(self, task):
        """Call the proper handler for the given C{task}."""
        message = task.data
        if message["type"] == "release-upgrade":
            return self.handle_release_upgrade(message)

    def handle_release_upgrade(self, message):
        """Fetch the upgrade-tool, verify it and run it.

        @param message: A message of type C{"release-upgrade"}.
        """
        target_code_name = message["code-name"]
        operation_id = message["operation-id"]
        lsb_release_info = parse_lsb_release(self.lsb_release_filename)
        current_code_name = lsb_release_info["code-name"]

        if target_code_name == current_code_name:
            message = self.make_operation_result_message(
                operation_id, FAILED,
                "The system is already running %s." % target_code_name, 1)
            logging.info("Queuing message with release upgrade failure to "
                         "exchange urgently.")
            return self._broker.send_message(message, self._session_id, True)

        tarball_url = message["upgrade-tool-tarball-url"]
        signature_url = message["upgrade-tool-signature-url"]
        allow_third_party = message.get("allow-third-party", False)
        debug = message.get("debug", False)
        mode = None
        if current_code_name == "dapper":
            # On Dapper the upgrade tool must be passed "--mode server"
            # when run on a server system. As there is no simple and
            # reliable way to detect if a system is a desktop one, and as
            # the desktop edition is no longer supported, we default to server
            # mode.
            mode = "server"
        directory = self._config.upgrade_tool_directory
        tarball_filename = url_to_filename(tarball_url,
                                           directory=directory)
        signature_filename = url_to_filename(signature_url,
                                             directory=directory)

        result = self.fetch(tarball_url, signature_url)
        result.addCallback(lambda x: self.verify(tarball_filename,
                                                 signature_filename))
        result.addCallback(lambda x: self.extract(tarball_filename))
        result.addCallback(lambda x: self.tweak(current_code_name))
        result.addCallback(lambda x: self.upgrade(
            target_code_name, operation_id,
            allow_third_party=allow_third_party, debug=debug, mode=mode))
        result.addCallback(lambda x: self.finish())
        result.addErrback(self.abort, operation_id)
        return result

    def fetch(self, tarball_url, signature_url):
        """Fetch the upgrade-tool files.

        @param tarball_url: The upgrade-tool tarball URL.
        @param signature_url: The upgrade-tool signature URL.
        """
        if not os.path.exists(self._config.upgrade_tool_directory):
            os.mkdir(self._config.upgrade_tool_directory)

        result = fetch_to_files([tarball_url, signature_url],
                                self._config.upgrade_tool_directory,
                                logger=logging.warning)

        def log_success(ignored):
            logging.info("Successfully fetched upgrade-tool files")

        def log_failure(failure):
            logging.warning("Couldn't fetch all upgrade-tool files")
            return failure

        result.addCallback(log_success)
        result.addErrback(log_failure)
        return result

    def verify(self, tarball_filename, signature_filename):
        """Verify the upgrade-tool tarball against its signature.

        @param tarball_filename: The filename of the upgrade-tool tarball.
        @param signature_filename: The filename of the tarball signature.
        """
        result = gpg_verify(tarball_filename, signature_filename)

        def log_success(ignored):
            logging.info("Successfully verified upgrade-tool tarball")

        def log_failure(failure):
            logging.warning("Invalid signature for upgrade-tool tarball: %s"
                            % str(failure.value))
            return failure

        result.addCallback(log_success)
        result.addErrback(log_failure)
        return result

    def extract(self, tarball_filename):
        """Extract the upgrade-tool tarball.

        @param tarball_filename: The filename of the upgrade-tool tarball.
        """
        tf = tarfile.open(tarball_filename, "r:gz")
        for member in tf.getmembers():
            tf.extract(member, path=self._config.upgrade_tool_directory)
        return succeed(None)

    def tweak(self, current_code_name):
        """Tweak the files of the extracted tarballs to workaround known bugs.

        @param current_code_name: The code-name of the current release.
        """
        upgrade_tool_directory = self._config.upgrade_tool_directory

        if current_code_name == "dapper":
            config_filename = os.path.join(upgrade_tool_directory,
                                           "DistUpgrade.cfg.dapper")
            config = ConfigParser.ConfigParser()
            config.read(config_filename)

            # Fix a bug in the DistUpgrade.cfg.dapper file contained in
            # the upgrade tool tarball
            if not config.has_section("NonInteractive"):
                config.add_section("NonInteractive")
                config.set("NonInteractive", "ForceOverwrite", "no")

            # Workaround for Bug #174148, which prevents dbus from restarting
            # after a dapper->hardy upgrade
            if not config.has_section("Distro"):
                config.add_section("Distro")
            if not config.has_option("Distro", "PostInstallScripts"):
                config.set("Distro", "PostInstallScripts", "./dbus.sh")
            else:
                scripts = config.get("Distro", "PostInstallScripts")
                scripts += ", ./dbus.sh"
                config.set("Distro", "PostInstallScripts", scripts)

            # Write config changes to disk
            fd = open(config_filename, "w")
            config.write(fd)
            fd.close()

            # Generate the post-install script that starts DBus
            dbus_sh_filename = os.path.join(upgrade_tool_directory,
                                            "dbus.sh")
            fd = open(dbus_sh_filename, "w")
            fd.write("#!/bin/sh\n"
                     "/etc/init.d/dbus start\n"
                     "sleep 10\n")
            fd.close()
            os.chmod(dbus_sh_filename, 0755)

        # On some releases the upgrade-tool doesn't support the allow third
        # party environment variable, so this trick is needed to make it
        # possible to upgrade against testing client packages from the
        # Landscape PPA
        mirrors_filename = os.path.join(upgrade_tool_directory,
                                        "mirrors.cfg")
        fd = open(mirrors_filename, "a")
        fd.write(self.landscape_ppa_url + "\n")
        fd.close()

        return succeed(None)

    def make_operation_result_text(self, out, err):
        """Return the operation result text to be sent to the server.

        @param out: The standard output of the upgrade-tool process.
        @param err: The standard error of the upgrade-tool process.
        @return: A text aggregating the process output, error and log files.
        """
        buf = cStringIO.StringIO()

        for label, content in [("output", out), ("error", err)]:
            if content:
                buf.write("=== Standard %s ===\n\n%s\n\n" % (label, content))

        for basename in sorted(os.listdir(self.logs_directory)):
            if not basename.endswith(".log"):
                continue
            filename = os.path.join(self.logs_directory, basename)
            content = read_file(filename, - self.logs_limit)
            buf.write("=== %s ===\n\n%s\n\n" % (basename, content))

        return buf.getvalue()

    def upgrade(self, code_name, operation_id, allow_third_party=False,
                debug=False, mode=None):
        """Run the upgrade-tool command and send a report of the results.

        @param code_name: The code-name of the release to upgrade to.
        @param operation_id: The activity id for this task.
        @param allow_third_party: Whether to enable non-official APT repo.
        @param debug: Whether to turn on debug level logging.
        @param mode: Optionally, the mode to run the upgrade-tool as. It
            can be "server" or "desktop", and it's relevant only for dapper.
        """
        upgrade_tool_directory = self._config.upgrade_tool_directory
        upgrade_tool_filename = os.path.join(upgrade_tool_directory, code_name)
        args = ["--frontend", "DistUpgradeViewNonInteractive"]
        if mode:
            args.extend(["--mode", mode])
        env = os.environ.copy()
        if allow_third_party:
            env["RELEASE_UPRADER_ALLOW_THIRD_PARTY"] = "True"
        if debug:
            env["DEBUG_UPDATE_MANAGER"] = "True"

        result = spawn_process(upgrade_tool_filename, args=args, env=env,
                               path=upgrade_tool_directory, wait_pipes=False)

        def send_operation_result((out, err, code)):
            if code == 0:
                status = SUCCEEDED
            else:
                status = FAILED
            text = self.make_operation_result_text(out, err)
            message = self.make_operation_result_message(operation_id, status,
                                                         text, code)
            logging.info("Queuing message with release upgrade results to "
                         "exchange urgently.")
            return self._broker.send_message(message, self._session_id, True)

        result.addCallback(send_operation_result)
        return result

    def finish(self):
        """Clean-up the upgrade-tool files and report about package changes."""
        shutil.rmtree(self._config.upgrade_tool_directory)

        if os.getuid() == 0:
            uid = pwd.getpwnam("landscape").pw_uid
            gid = grp.getgrnam("landscape").gr_gid
        else:
            uid = None
            gid = None

        reporter = find_reporter_command()

        # Force an apt-update run, because the sources.list has changed
        args = ["--force-apt-update"]

        if self._config.config is not None:
            args.append("--config=%s" % self._config.config)

        return spawn_process(reporter, args=args, uid=uid, gid=gid,
                             path=os.getcwd(), env=os.environ)

    def abort(self, failure, operation_id):
        """Abort the task reporting details about the failure."""

        message = self.make_operation_result_message(
            operation_id, FAILED, "%s" % str(failure.value), 1)

        logging.info("Queuing message with release upgrade failure to "
                     "exchange urgently.")

        return self._broker.send_message(message, self._session_id, True)

    @staticmethod
    def find_command():
        return find_release_upgrader_command()


def find_release_upgrader_command():
    """Return the path to the landscape-release-upgrader script."""
    dirname = os.path.dirname(os.path.abspath(sys.argv[0]))
    return os.path.join(dirname, "landscape-release-upgrader")


def main(args):
    if os.getpgrp() != os.getpid():
        os.setsid()
    return run_task_handler(ReleaseUpgrader, args)



import array
from cStringIO import StringIO
import socket
from subprocess import Popen, PIPE
from landscape.tests.helpers import LandscapeTest

from landscape.lib.network import (
    get_network_traffic, get_active_device_info, get_active_interfaces,
    get_fqdn, get_network_interface_speed)
from landscape.tests.mocker import ANY


class NetworkInfoTest(LandscapeTest):

    def test_get_active_device_info(self):
        """
        Device info returns a sequence of information about active
        network devices, compare and verify the output against
        that returned by ifconfig.
        """
        mock_get_network_interface_speed = self.mocker.replace(
            get_network_interface_speed)
        mock_get_network_interface_speed(ANY, ANY)
        self.mocker.result((100, True))
        self.mocker.count(min=1, max=None)
        self.mocker.replay()

        device_info = get_active_device_info()
        result = Popen(["/sbin/ifconfig"], stdout=PIPE).communicate()[0]
        interface_blocks = dict(
            [(block.split()[0], block.upper()) for block in
             filter(None, result.split("\n\n"))])

        for device in device_info:
            if device["mac_address"] == "00:00:00:00:00:00":
                continue
            self.assertTrue(device["interface"] in result)
            block = interface_blocks[device["interface"]]
            self.assertTrue(device["netmask"] in block)
            self.assertIn(device["ip_address"], block)
            self.assertIn(device["mac_address"].upper(), block)
            self.assertIn(device["broadcast_address"], block)
            flags = device["flags"]
            if flags & 1:
                self.assertIn("UP", block)
            if flags & 2:
                self.assertIn("BROADCAST", block)
            if flags & 64:
                self.assertIn("RUNNING", block)
            if flags & 4096:
                self.assertIn("MULTICAST", block)
            self.assertEqual(100, device["speed"])
            self.assertEqual(True, device["duplex"])

    def test_skip_loopback(self):
        """The C{lo} interface is not reported by L{get_active_device_info}."""
        device_info = get_active_device_info()
        interfaces = [i["interface"] for i in device_info]
        self.assertNotIn("lo", interfaces)

    def test_skip_vlan(self):
        """VLAN interfaces are not reported by L{get_active_device_info}."""
        mock_get_active_interfaces = self.mocker.replace(get_active_interfaces)
        mock_get_active_interfaces(ANY)
        self.mocker.passthrough(
            result_callback=lambda result: list(result) + ["eth0.1"])
        self.mocker.replay()
        device_info = get_active_device_info()
        interfaces = [i["interface"] for i in device_info]
        self.assertNotIn("eth0.1", interfaces)

    def test_skip_alias(self):
        """Interface aliases are not reported by L{get_active_device_info}."""
        mock_get_active_interfaces = self.mocker.replace(get_active_interfaces)
        mock_get_active_interfaces(ANY)
        self.mocker.passthrough(
            result_callback=lambda result: list(result) + ["eth0:foo"])
        self.mocker.replay()
        device_info = get_active_device_info()
        interfaces = [i["interface"] for i in device_info]
        self.assertNotIn("eth0:foo", interfaces)

    def test_duplicate_network_interfaces(self):
        """
        L{get_active_interfaces} doesn't return duplicate network interfaces.
        The call to C{fcntl.ioctl} might return the same interface several
        times, so we make sure to clean it up.
        """
        import landscape.lib.network
        original_struct_size = landscape.lib.network.IF_STRUCT_SIZE
        landscape.lib.network.IF_STRUCT_SIZE = 40
        self.addCleanup(
            setattr, landscape.lib.network, "IF_STRUCT_SIZE",
            original_struct_size)
        # This is a fake response observed to return the same interface several
        # times (here, br1:priv)
        response = (
            "lo\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02"
            "\x00\x00\x00\x7f\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00eth1:pub\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x02\x00\x00\x00\xc8\xb4\xc4.\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00br1:metadata\x00\x00\x00\x00\x02"
            "\x00\x00\x00\xa9\xfe\xa9\xfe\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00br1:0\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x02\x00\x00\x00\xc9\x19\x1f\x1d\x00\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00br1\x00\x00\x00\x00"
            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\xc0\xa8d"
            "\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00br1:priv\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\xac"
            "\x13\x02\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            "\x00\x00\x00br1:priv\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00"
            "\x00\xac\x13\x02A")

        fake_array = array.array("B", response + "\0" * 4855)
        mock_array = self.mocker.replace("array")
        mock_array.array("B", ANY)
        self.mocker.result(fake_array)

        mock_ioctl = self.mocker.replace("fcntl")
        mock_ioctl.ioctl(ANY, ANY, ANY)
        self.mocker.result(0)

        mock_unpack = self.mocker.replace("struct")
        mock_unpack.unpack("iL", ANY)
        self.mocker.result((280, 38643456))

        self.mocker.replay()

        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP)
        self.assertEqual(
            ["lo", "eth1:pub", "br1:metadata", "br1:0", "br1", "br1:priv"],
            list(get_active_interfaces(sock)))

    def test_get_network_traffic(self):
        """
        Network traffic is assessed via reading /proc/net/dev, verify
        the parsed output against a known sample.
        """
        open_mock = self.mocker.replace("__builtin__.open")
        open_mock("/proc/net/dev", "r")
        self.mocker.result(StringIO(test_proc_net_dev_output))
        self.mocker.replay()
        traffic = get_network_traffic()
        self.assertEqual(traffic, test_proc_net_dev_parsed)


#exact output of cat /proc/net/dev snapshot with line continuations for pep8
test_proc_net_dev_output = """\
Inter-|   Receive                                                |  Transmit
 face |bytes    packets errs drop fifo frame compressed multicast|bytes    \
 packets errs drop fifo colls carrier compressed
    lo:3272627934 3321049    0    0    0     0          0         0 3272627934\
    3321049    0    0    0     0       0          0
  eth0: 6063748   12539    0    0    0     0          0        62  2279693\
  12579    0    0    0    19       0          0
"""

test_proc_net_dev_parsed = {
    "lo": {"recv_bytes": 3272627934,
          "recv_packets": 3321049,
          "recv_errs": 0,
          "recv_drop": 0,
          "recv_fifo": 0,
          "recv_frame": 0,
          "recv_compressed": 0,
          "recv_multicast": 0,
          "send_bytes": 3272627934,
          "send_packets": 3321049,
          "send_errs": 0,
          "send_drop": 0,
          "send_fifo": 0,
          "send_colls": 0,
          "send_carrier": 0,
          "send_compressed": 0},
    "eth0": {"recv_bytes": 6063748,
            "recv_packets": 12539,
            "recv_errs": 0,
            "recv_drop": 0,
            "recv_fifo": 0,
            "recv_frame": 0,
            "recv_compressed": 0,
            "recv_multicast": 62,
            "send_bytes": 2279693,
            "send_packets": 12579,
            "send_errs": 0,
            "send_drop": 0,
            "send_fifo": 0,
            "send_colls": 19,
            "send_carrier": 0,
            "send_compressed": 0}}


class FQDNTest(LandscapeTest):

    def test_default_fqdn(self):
        """
        C{get_fqdn} returns the output of C{socket.getfqdn} if it returns
        something sensible.
        """
        self.addCleanup(setattr, socket, "getfqdn", socket.getfqdn)
        socket.getfqdn = lambda: "foo.bar"
        self.assertEqual("foo.bar", get_fqdn())

    def test_getaddrinfo_fallback(self):
        """
        C{get_fqdn} falls back to C{socket.getaddrinfo} with the
        C{AI_CANONNAME} flag if C{socket.getfqdn} returns a local hostname.
        """
        self.addCleanup(setattr, socket, "getfqdn", socket.getfqdn)
        socket.getfqdn = lambda: "localhost6.localdomain6"
        self.assertNotIn("localhost", get_fqdn())


class NetworkInterfaceSpeedTest(LandscapeTest):

    def test_get_network_interface_speed(self):
        """
        The link speed is reported as unpacked from the ioctl() call.
        """
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP)
        # ioctl always succeeds
        mock_ioctl = self.mocker.replace("fcntl")
        mock_ioctl.ioctl(ANY, ANY, ANY)
        self.mocker.result(0)

        mock_unpack = self.mocker.replace("struct")
        mock_unpack.unpack("12xHB28x", ANY)
        self.mocker.result((100, False))

        self.mocker.replay()

        self.assertEqual((100, False),
                         get_network_interface_speed(sock, "eth0"))

    def test_get_network_interface_speed_unplugged(self):
        """
        The link speed for an unplugged interface is reported as 0.
        """
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP)
        # ioctl always succeeds
        mock_ioctl = self.mocker.replace("fcntl")
        mock_ioctl.ioctl(ANY, ANY, ANY)
        self.mocker.result((0, False))

        mock_unpack = self.mocker.replace("struct")
        mock_unpack.unpack("12xHB28x", ANY)
        self.mocker.result((65535, False))

        self.mocker.replay()

        self.assertEqual((0, False), get_network_interface_speed(sock, "eth0"))

    def test_get_network_interface_speed_not_supported(self):
        """
        Some drivers do not report the needed interface speed. In this case
        an C{IOError} with errno 95 is raised ("Operation not supported").
        If such an error is rasied, report the speed as -1.
        """
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP)
        theerror = IOError()
        theerror.errno = 95
        theerror.message = "Operation not supported"

        # ioctl always raises
        mock_ioctl = self.mocker.replace("fcntl")
        mock_ioctl.ioctl(ANY, ANY, ANY)
        self.mocker.throw(theerror)

        self.mocker.replay()

        self.assertEqual((-1, False),
                         get_network_interface_speed(sock, "eth0"))

    def test_get_network_interface_speed_not_permitted(self):
        """
        In some cases (lucid seems to be affected), the ioctl() call is not
        allowed for non-root users. In that case we intercept the error and
        not report the network speed.
        """
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP)
        theerror = IOError()
        theerror.errno = 1
        theerror.message = "Operation not permitted"

        # ioctl always raises
        mock_ioctl = self.mocker.replace("fcntl")
        mock_ioctl.ioctl(ANY, ANY, ANY)
        self.mocker.throw(theerror)

        self.mocker.replay()

        self.assertEqual((-1, False),
                         get_network_interface_speed(sock, "eth0"))

    def test_get_network_interface_speed_other_io_error(self):
        """
        In case we get an IOError that is not "Operation not permitted", the
        exception should be raised.
        """
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_IP)
        theerror = IOError()
        theerror.errno = 999
        theerror.message = "Whatever"

        # ioctl always raises
        mock_ioctl = self.mocker.replace("fcntl")
        mock_ioctl.ioctl(ANY, ANY, ANY)
        self.mocker.throw(theerror)

        self.mocker.replay()

        self.assertRaises(IOError, get_network_interface_speed, sock, "eth0")



"""
Network introspection utilities using ioctl and the /proc filesystem.
"""
import array
import fcntl
import socket
import struct
import errno
import logging

__all__ = ["get_active_device_info", "get_network_traffic", "is_64"]

# from header /usr/include/bits/ioctls.h
SIOCGIFCONF = 0x8912
SIOCGIFFLAGS = 0x8913
SIOCGIFNETMASK = 0x891b
SIOCGIFBRDADDR = 0x8919
SIOCGIFADDR = 0x8915
SIOCGIFHWADDR = 0x8927

SIOCETHTOOL = 0x8946  # As defined in include/uapi/linux/sockios.h
ETHTOOL_GSET = 0x00000001  # Get status command.

# struct definition from header /usr/include/net/if.h
# the struct size varies according to the platform arch size
# a minimal c program was used to determine the size of the
# struct, standard headers removed for brevity.
"""
#include <linux/if.h>
int main() {
  printf("Size of struct %lu\n", sizeof(struct ifreq));
}
"""

IF_STRUCT_SIZE_32 = 32
IF_STRUCT_SIZE_64 = 40


def is_64():
    """Returns C{True} if the platform is 64-bit, otherwise C{False}."""
    return struct.calcsize("l") == 8


# initialize the struct size as per the machine's architecture
IF_STRUCT_SIZE = is_64() and IF_STRUCT_SIZE_64 or IF_STRUCT_SIZE_32


def get_active_interfaces(sock):
    """Generator yields active network interface names.

    @param sock: a socket instance.
    """
    max_interfaces = 128

    # Setup an array to hold our response, and initialized to null strings.
    interfaces = array.array("B", "\0" * max_interfaces * IF_STRUCT_SIZE)
    buffer_size = interfaces.buffer_info()[0]
    packed_bytes = struct.pack(
        "iL", max_interfaces * IF_STRUCT_SIZE, buffer_size)

    byte_length = struct.unpack(
        "iL", fcntl.ioctl(sock.fileno(), SIOCGIFCONF, packed_bytes))[0]

    result = interfaces.tostring()

    # Generator over the interface names
    already_found = set()
    for index in range(0, byte_length, IF_STRUCT_SIZE):
        ifreq_struct = result[index:index + IF_STRUCT_SIZE]
        interface_name = ifreq_struct[:ifreq_struct.index("\0")]
        if interface_name not in already_found:
            already_found.add(interface_name)
            yield interface_name


def get_broadcast_address(sock, interface):
    """Return the broadcast address associated to an interface.

    @param sock: a socket instance.
    @param interface: The name of the interface.
    """
    return socket.inet_ntoa(fcntl.ioctl(
        sock.fileno(),
        SIOCGIFBRDADDR,
        struct.pack("256s", interface[:15]))[20:24])


def get_netmask(sock, interface):
    """Return the network mask associated to an interface.

    @param sock: a socket instance.
    @param interface: The name of the interface.
    """
    return socket.inet_ntoa(fcntl.ioctl(
        sock.fileno(),
        SIOCGIFNETMASK,
        struct.pack("256s", interface[:15]))[20:24])


def get_ip_address(sock, interface):
    """Return the IP address associated to the interface.

    @param sock: a socket instance.
    @param interface: The name of the interface.
    """
    return socket.inet_ntoa(fcntl.ioctl(
        sock.fileno(),
        SIOCGIFADDR,
        struct.pack("256s", interface[:15]))[20:24])


def get_mac_address(sock, interface):
    """
    Return the hardware MAC address for an interface in human friendly form,
    ie. six colon separated groups of two hexadecimal digits.

    @param sock: a socket instance.
    @param interface: The name of the interface.
    """
    mac_address = fcntl.ioctl(
        sock.fileno(), SIOCGIFHWADDR, struct.pack("256s", interface[:15]))
    return "".join(["%02x:" % ord(char) for char in mac_address[18:24]])[:-1]


def get_flags(sock, interface):
    """Return the integer value of the interface flags for the given interface.

    @param sock: a socket instance.
    @param interface: The name of the interface.
    @see /usr/include/linux/if.h for the meaning of the flags.
    """
    data = fcntl.ioctl(
        sock.fileno(), SIOCGIFFLAGS, struct.pack("256s", interface[:15]))
    return struct.unpack("H", data[16:18])[0]


def get_active_device_info(skipped_interfaces=("lo",),
                           skip_vlan=True, skip_alias=True):
    """
    Returns a dictionary containing information on each active network
    interface present on a machine.
    """
    results = []
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM,
                             socket.IPPROTO_IP)
        for interface in get_active_interfaces(sock):
            if interface in skipped_interfaces:
                continue
            if skip_vlan and "." in interface:
                continue
            if skip_alias and ":" in interface:
                continue
            interface_info = {"interface": interface}
            interface_info["ip_address"] = get_ip_address(sock, interface)
            interface_info["mac_address"] = get_mac_address(sock, interface)
            interface_info["broadcast_address"] = get_broadcast_address(
                sock, interface)
            interface_info["netmask"] = get_netmask(sock, interface)
            interface_info["flags"] = get_flags(sock, interface)
            speed, duplex = get_network_interface_speed(sock, interface)
            interface_info["speed"] = speed
            interface_info["duplex"] = duplex
            results.append(interface_info)
    finally:
        del sock

    return results


def get_network_traffic(source_file="/proc/net/dev"):
    """
    Retrieves an array of information regarding the network activity per
    network interface.
    """
    netdev = open(source_file, "r")
    lines = netdev.readlines()
    netdev.close()

    # Parse out the column headers as keys.
    _, receive_columns, transmit_columns = lines[1].split("|")
    columns = ["recv_%s" % column for column in receive_columns.split()]
    columns.extend(["send_%s" % column for column in transmit_columns.split()])

    # Parse out the network devices.
    devices = {}
    for line in lines[2:]:
        if not ":" in line:
            continue
        device, data = line.split(":")
        device = device.strip()
        devices[device] = dict(zip(columns, map(long, data.split())))
    return devices


def get_fqdn():
    """
    Return the current fqdn of the machine, trying hard to return a meaningful
    name.

    In particular, it means working against a NetworkManager bug which seems to
    make C{getfqdn} return localhost6.localdomain6 for machine without a domain
    since Maverick.
    """
    fqdn = socket.getfqdn()
    if "localhost" in fqdn:
        # Try the heavy artillery
        fqdn = socket.getaddrinfo(socket.gethostname(), None, socket.AF_INET,
                                  socket.SOCK_DGRAM, socket.IPPROTO_IP,
                                  socket.AI_CANONNAME)[0][3]
        if "localhost" in fqdn:
            # Another fallback
            fqdn = socket.gethostname()
    return fqdn


def get_network_interface_speed(sock, interface_name):
    """
    Return the ethernet device's advertised link speed.

    The return value can be one of:
        * 10, 100, 1000, 2500, 10000: The interface speed in Mbps
        * -1: The interface does not support querying for max speed, such as
          virtio devices for instance.
        * 0: The cable is not connected to the interface. We cannot measure
          interface speed, but could if it was plugged in.
    """
    cmd_struct = struct.pack('I39s', ETHTOOL_GSET, '\x00' * 39)
    status_cmd = array.array('B', cmd_struct)
    packed = struct.pack('16sP', interface_name, status_cmd.buffer_info()[0])

    speed = -1
    try:
        fcntl.ioctl(sock, SIOCETHTOOL, packed)  # Status ioctl() call
        res = status_cmd.tostring()
        speed, duplex = struct.unpack('12xHB28x', res)
    except IOError as e:
        if e.errno == errno.EPERM:
            logging.warn("Could not determine network interface speed, "
                         "operation not permitted.")
        elif e.errno != errno.EOPNOTSUPP:
            raise e
        speed = -1
        duplex = False

    # Drivers apparently report speed as 65535 when the link is not available
    # (cable unplugged for example).
    if speed == 65535:
        speed = 0

    # The drivers report "duplex" to be 255 when the information is not
    # available. We'll just assume it's False in that case.
    if duplex == 255:
        duplex = False
    duplex = bool(duplex)

    return speed, duplex



from landscape import SERVER_API, CLIENT_API
from landscape.lib.persist import Persist
from landscape.lib.hashlib import md5
from landscape.lib.fetch import fetch_async
from landscape.schema import Message, Int
from landscape.broker.config import BrokerConfiguration
from landscape.broker.exchange import get_accepted_types_diff, MessageExchange
from landscape.broker.transport import FakeTransport
from landscape.broker.store import MessageStore
from landscape.broker.ping import Pinger
from landscape.broker.registration import RegistrationHandler
from landscape.tests.helpers import (LandscapeTest, DEFAULT_ACCEPTED_TYPES)
from landscape.tests.mocker import MATCH
from landscape.broker.tests.helpers import ExchangeHelper
from landscape.broker.server import BrokerServer


class RaisingTransport(object):

    def get_url(self2):
        return ""

    def exchange(self2, *args):
        raise RuntimeError("Failed to communicate.")


class MessageExchangeTest(LandscapeTest):

    helpers = [ExchangeHelper]

    def setUp(self):
        super(MessageExchangeTest, self).setUp()
        self.mstore.add_schema(Message("empty", {}))
        self.mstore.add_schema(Message("data", {"data": Int()}))
        self.mstore.add_schema(Message("holdme", {}))
        self.identity.secure_id = 'needs-to-be-set-for-tests-to-pass'

    def wait_for_exchange(self, urgent=False, factor=1, delta=0):
        if urgent:
            seconds = self.config.urgent_exchange_interval
        else:
            seconds = self.config.exchange_interval
        self.reactor.advance(seconds * factor + delta)

    def test_resynchronize_causes_urgent_exchange(self):
        """
        A 'resynchronize-clients' messages causes an urgent exchange
        to be scheduled.
        """
        self.assertFalse(self.exchanger.is_urgent())
        self.reactor.fire("resynchronize-clients")
        self.assertTrue(self.exchanger.is_urgent())

    def test_that_resynchronize_drops_session_ids(self):
        """
        When a resynchronisation event occurs with global scope all existing
        session IDs are expired, so any new messages being sent with those IDs
        will be discarded.
        """
        broker = BrokerServer(self.config, self.reactor,
                              self.exchanger, None,
                              self.mstore, None)

        disk_session_id = self.mstore.get_session_id(scope="disk")
        package_session_id = self.mstore.get_session_id(scope="package")
        self.mstore.set_accepted_types(["empty"])
        global_scope = []
        self.reactor.fire("resynchronize-clients", global_scope)
        broker.send_message({"type": "empty"}, disk_session_id)
        broker.send_message({"type": "empty"}, package_session_id)
        self.exchanger.exchange()
        messages = self.transport.payloads[0]["messages"]
        self.assertMessages(messages, [])

    def test_that_resynchronize_drops_scoped_session_ids_only(self):
        """
        When a resynchronisation event occurs with a scope existing session IDs
        for that scope are expired, all other session IDs are unaffected.
        """
        broker = BrokerServer(self.config, self.reactor,
                              self.exchanger, None,
                              self.mstore, None)

        disk_session_id = self.mstore.get_session_id(scope="disk")
        package_session_id = self.mstore.get_session_id(scope="package")
        self.mstore.set_accepted_types(["empty"])
        disk_scope = ["disk"]
        self.reactor.fire("resynchronize-clients", disk_scope)
        broker.send_message({"type": "empty"}, disk_session_id)
        broker.send_message({"type": "empty"}, package_session_id)
        self.exchanger.exchange()
        messages = self.transport.payloads[0]["messages"]
        self.assertMessages(messages, [{"type": "empty"}])

    def test_that_resynchronize_clears_message_blackhole(self):
        """
        When a resynchronisation event occurs the block on new messages
        being stored is lifted.
        """
        self.reactor.fire("resynchronize-clients", [])
        persist = Persist(filename=self.persist_filename)
        self.assertFalse(persist.get("blackhole-messages"))

    def test_send(self):
        """
        The send method should cause a message to show up in the next exchange.
        """
        self.mstore.set_accepted_types(["empty"])
        self.exchanger.send({"type": "empty"})
        self.exchanger.exchange()
        self.assertEqual(len(self.transport.payloads), 1)
        messages = self.transport.payloads[0]["messages"]
        self.assertEqual(messages, [{"type": "empty",
                                     "timestamp": 0,
                                     "api": SERVER_API}])

    def test_send_urgent(self):
        """
        Sending a message with the urgent flag should schedule an
        urgent exchange.
        """
        self.mstore.set_accepted_types(["empty"])
        self.exchanger.send({"type": "empty"}, urgent=True)
        self.wait_for_exchange(urgent=True)
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [{"type": "empty"}])

    def test_send_urgent_wont_reschedule(self):
        """
        If an urgent exchange is already scheduled, adding another
        urgent message shouldn't reschedule the exchange forward.
        """
        self.mstore.set_accepted_types(["empty"])
        self.exchanger.send({"type": "empty"}, urgent=True)
        self.wait_for_exchange(urgent=True, factor=0.5)
        self.exchanger.send({"type": "empty"}, urgent=True)
        self.wait_for_exchange(urgent=True, factor=0.5)
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [{"type": "empty"}, {"type": "empty"}])

    def test_send_returns_message_id(self):
        """
        The send method should return the message id, as returned by add().
        """
        self.mstore.set_accepted_types(["empty"])
        message_id = self.exchanger.send({"type": "empty"})
        self.assertTrue(self.mstore.is_pending(message_id))
        self.mstore.add_pending_offset(1)
        self.assertFalse(self.mstore.is_pending(message_id))

    def test_wb_include_accepted_types(self):
        """
        Every payload from the client needs to specify an ID which
        represents the types that we think the server wants.
        """
        payload = self.exchanger._make_payload()
        self.assertIn("accepted-types", payload)
        self.assertEqual(payload["accepted-types"], md5("").digest())

    def test_handle_message_sets_accepted_types(self):
        """
        An incoming "accepted-types" message should set the accepted
        types.
        """
        self.exchanger.handle_message(
            {"type": "accepted-types", "types": ["foo"]})
        self.assertEqual(self.mstore.get_accepted_types(), ["foo"])

    def test_message_type_acceptance_changed_event(self):
        stash = []

        def callback(type, accepted):
            stash.append((type, accepted))
        self.reactor.call_on("message-type-acceptance-changed", callback)
        self.exchanger.handle_message(
            {"type": "accepted-types", "types": ["a", "b"]})
        self.exchanger.handle_message(
            {"type": "accepted-types", "types": ["b", "c"]})
        self.assertEqual(stash, [("a", True), ("b", True),
                                 ("a", False), ("c", True)])

    def test_wb_accepted_types_roundtrip(self):
        """
        Telling the client to set the accepted types with a message
        should affect its future payloads.
        """
        self.exchanger.handle_message(
            {"type": "accepted-types", "types": ["ack", "bar"]})
        payload = self.exchanger._make_payload()
        self.assertIn("accepted-types", payload)
        self.assertEqual(payload["accepted-types"],
                         md5("ack;bar").digest())

    def test_accepted_types_causes_urgent_if_held_messages_exist(self):
        """
        If an accepted-types message makes available a type for which we
        have a held message, an urgent exchange should occur.
        """
        self.exchanger.send({"type": "holdme"})
        self.assertEqual(self.mstore.get_pending_messages(), [])
        self.exchanger.handle_message(
            {"type": "accepted-types", "types": ["holdme"]})
        self.wait_for_exchange(urgent=True)
        self.assertEqual(len(self.transport.payloads), 1)
        self.assertMessages(self.transport.payloads[0]["messages"],
                            [{"type": "holdme"}])

    def test_accepted_types_no_urgent_without_held(self):
        """
        If an accepted-types message does *not* "unhold" any exist messages,
        then no urgent exchange should occur.
        """
        self.exchanger.send({"type": "holdme"})
        self.assertEqual(self.transport.payloads, [])
        self.reactor.fire("message",
                          {"type": "accepted-types", "types": ["irrelevant"]})
        self.assertEqual(len(self.transport.payloads), 0)

    def test_sequence_is_committed_immediately(self):
        """
        The MessageStore should be committed by the MessageExchange as soon as
        possible after setting the pending offset and sequence.
        """
        self.mstore.set_accepted_types(["empty"])

        # We'll check that the message store has been saved by the time a
        # message handler gets called.
        self.transport.responses.append([{"type": "inbound"}])
        self.exchanger.send({"type": "empty"})

        handled = []

        def handler(message):
            persist = Persist(filename=self.persist_filename)
            store = MessageStore(persist, self.config.message_store_path)
            self.assertEqual(store.get_pending_offset(), 1)
            self.assertEqual(store.get_sequence(), 1)
            handled.append(True)

        self.exchanger.register_message("inbound", handler)
        self.exchanger.exchange()
        self.assertEqual(handled, [True], self.logfile.getvalue())

    def test_messages_from_server_commit(self):
        """
        The Exchange should commit the message store after processing each
        message.
        """
        self.transport.responses.append([{"type": "inbound"}] * 3)
        handled = []
        self.message_counter = 0

        def handler(message):
            Persist(filename=self.persist_filename)
            store = MessageStore(self.persist, self.config.message_store_path)
            self.assertEqual(store.get_server_sequence(),
                             self.message_counter)
            self.message_counter += 1
            handled.append(True)

        self.exchanger.register_message("inbound", handler)
        self.exchanger.exchange()
        self.assertEqual(handled, [True] * 3, self.logfile.getvalue())

    def test_messages_from_server_causing_urgent_exchanges(self):
        """
        If a message from the server causes an urgent message to be
        queued, an urgent exchange should happen again after the
        running exchange.
        """
        self.transport.responses.append([{"type": "foobar"}])
        self.mstore.set_accepted_types(["empty"])

        def handler(message):
            self.exchanger.send({"type": "empty"}, urgent=True)

        self.exchanger.register_message("foobar", handler)

        self.exchanger.exchange()

        self.assertEqual(len(self.transport.payloads), 1)

        self.wait_for_exchange(urgent=True)

        self.assertEqual(len(self.transport.payloads), 2)
        self.assertMessages(self.transport.payloads[1]["messages"],
                            [{"type": "empty"}])

    def test_server_expects_older_messages(self):
        """
        If the server expects an old message, the exchanger should be
        marked as urgent.
        """
        self.mstore.set_accepted_types(["data"])
        self.mstore.add({"type": "data", "data": 0})
        self.mstore.add({"type": "data", "data": 1})
        self.exchanger.exchange()
        self.assertEqual(self.mstore.get_sequence(), 2)

        self.mstore.add({"type": "data", "data": 2})
        self.mstore.add({"type": "data", "data": 3})

        # next one, server will respond with 1!
        def desynched_send_data(*args, **kwargs):
            self.transport.next_expected_sequence = 1
            return {"next-expected-sequence": 1}

        self.transport.exchange = desynched_send_data
        self.exchanger.exchange()
        self.assertEqual(self.mstore.get_sequence(), 1)
        del self.transport.exchange

        exchanged = []

        def exchange_callback():
            exchanged.append(True)

        self.reactor.call_on("exchange-done", exchange_callback)
        self.wait_for_exchange(urgent=True)
        self.assertEqual(exchanged, [True])

        payload = self.transport.payloads[-1]
        self.assertMessages(payload["messages"],
                            [{"type": "data", "data": 1},
                             {"type": "data", "data": 2},
                             {"type": "data", "data": 3}])
        self.assertEqual(payload["sequence"], 1)
        self.assertEqual(payload["next-expected-sequence"], 0)

    def test_start_with_urgent_exchange(self):
        """
        Immediately after registration, an urgent exchange should be scheduled.
        """
        transport = FakeTransport()
        exchanger = MessageExchange(self.reactor, self.mstore, transport,
                                    self.identity, self.exchange_store,
                                    self.config)
        exchanger.start()
        self.wait_for_exchange(urgent=True)
        self.assertEqual(len(transport.payloads), 1)

    def test_reschedule_after_exchange(self):
        """
        Under normal operation, after an exchange has finished another
        exchange should be scheduled for after the normal delay.
        """
        self.exchanger.schedule_exchange(urgent=True)

        self.wait_for_exchange(urgent=True)
        self.assertEqual(len(self.transport.payloads), 1)

        self.wait_for_exchange()
        self.assertEqual(len(self.transport.payloads), 2)

        self.wait_for_exchange()
        self.assertEqual(len(self.transport.payloads), 3)

    def test_leave_urgent_exchange_mode_after_exchange(self):
        """
        After an urgent exchange, assuming no messages are left to be
        exchanged, urgent exchange should not remain scheduled.
        """
        self.mstore.set_accepted_types(["empty"])
        self.exchanger.send({"type": "empty"}, urgent=True)
        self.wait_for_exchange(urgent=True)
        self.assertEqual(len(self.transport.payloads), 1)
        self.wait_for_exchange(urgent=True)
        self.assertEqual(len(self.transport.payloads), 1)  # no change

    def test_successful_exchange_records_success(self):
        """
        When a successful exchange occurs, that success is recorded in the
        message store.
        """
        mock_message_store = self.mocker.proxy(self.mstore)
        mock_message_store.record_success(MATCH(lambda x: type(x) is int))
        self.mocker.result(None)
        self.mocker.replay()

        exchanger = MessageExchange(
            self.reactor, mock_message_store, self.transport,
            self.identity, self.exchange_store, self.config)
        exchanger.exchange()

    def test_ancient_causes_resynchronize(self):
        """
        If the server asks for messages that we no longer have, the message
        exchange plugin should send a message to the server indicating that a
        resynchronization is occuring and then fire a "resynchronize-clients"
        reactor message, so that plugins can generate new data -- if the server
        got out of synch with the client, then we're best off synchronizing
        everything back to it.
        """
        self.mstore.set_accepted_types(["empty", "data", "resynchronize"])
        # Do three generations of messages, so we "lose" the 0th message
        for i in range(3):
            self.mstore.add({"type": "empty"})
            self.exchanger.exchange()
        # the server loses some data
        self.transport.next_expected_sequence = 0

        def resynchronize(scopes=None):
            # We'll add a message to the message store here, since this is what
            # is commonly done in a resynchronize callback. This message added
            # should come AFTER the "resynchronize" message that is generated
            # by the exchange code itself.
            self.mstore.add({"type": "data", "data": 999})
        self.reactor.call_on("resynchronize-clients", resynchronize)

        # This exchange call will notice the server is asking for an old
        # message and fire the event:
        self.exchanger.exchange()
        self.assertMessages(self.mstore.get_pending_messages(),
                            [{"type": "empty"},
                             {"type": "resynchronize"},
                             {"type": "data", "data": 999}])

    def test_resynchronize_msg_causes_resynchronize_response_then_event(self):
        """
        If a message of type 'resynchronize' is received from the
        server, the exchanger should *first* send a 'resynchronize'
        message back to the server and *then* fire a
        'resynchronize-clients' event.
        """
        self.mstore.set_accepted_types(["empty", "resynchronize"])

        def resynchronized(scopes=None):
            self.mstore.add({"type": "empty"})
        self.reactor.call_on("resynchronize-clients", resynchronized)

        self.transport.responses.append([{"type": "resynchronize",
                                          "operation-id": 123}])
        self.exchanger.exchange()
        self.assertMessages(self.mstore.get_pending_messages(),
                            [{"type": "resynchronize",
                              "operation-id": 123},
                             {"type": "empty"}])

    def test_scopes_are_copied_from_incoming_resynchronize_messages(self):
        """
        If an incoming message of type 'reysnchronize' contains a 'scopes' key,
        then it's value is copied into the "resynchronize-clients" event.
        """
        fired_scopes = []
        self.mstore.set_accepted_types(["reysnchronize"])

        def resynchronized(scopes=None):
            fired_scopes.extend(scopes)

        self.reactor.call_on("resynchronize-clients", resynchronized)

        self.transport.responses.append([{"type": "resynchronize",
                                          "operation-id": 123,
                                          "scopes": ["disk", "users"]}])
        self.exchanger.exchange()
        self.assertEqual(["disk", "users"], fired_scopes)

    def test_no_urgency_when_server_expects_current_message(self):
        """
        When the message the server expects is the same as the first
        pending message sequence, the client should not go into urgent
        exchange mode.

        This means the server handler is likely blowing up and the client and
        the server are in a busy loop constantly asking for the same message,
        breaking, setting urgent exchange mode, sending the same message and
        then breaking in a fast loop.  In this case, urgent exchange mode
        should not be set. (bug #138135)
        """
        # We set the server sequence to some non-0 value to ensure that the
        # server and client sequences aren't the same to ensure the code is
        # looking at the correct sequence number. :(
        self.mstore.set_server_sequence(3300)
        self.mstore.set_accepted_types(["data"])
        self.mstore.add({"type": "data", "data": 0})

        def desynched_send_data(*args, **kwargs):
            self.transport.next_expected_sequence = 0
            return {"next-expected-sequence": 0}

        self.transport.exchange = desynched_send_data
        self.exchanger.exchange()

        self.assertEqual(self.mstore.get_sequence(), 0)
        del self.transport.exchange

        exchanged = []

        def exchange_callback():
            exchanged.append(True)

        self.reactor.call_on("exchange-done", exchange_callback)
        self.wait_for_exchange(urgent=True)
        self.assertEqual(exchanged, [])
        self.wait_for_exchange()
        self.assertEqual(exchanged, [True])

    def test_old_sequence_id_does_not_cause_resynchronize(self):
        resynchronized = []
        self.reactor.call_on("resynchronize",
                             lambda: resynchronized.append(True))

        self.mstore.set_accepted_types(["empty"])
        self.mstore.add({"type": "empty"})
        self.exchanger.exchange()
        # the server loses some data, but not too much
        self.transport.next_expected_sequence = 0

        self.exchanger.exchange()
        self.assertEqual(resynchronized, [])

    def test_per_api_payloads(self):
        """
        When sending messages to the server, the exchanger should split
        messages with different APIs in different payloads, and deliver
        them to the right API on the server.
        """
        types = ["a", "b", "c", "d", "e", "f"]
        self.mstore.set_accepted_types(types)
        for t in types:
            self.mstore.add_schema(Message(t, {}))

        self.exchanger.exchange()

        # No messages queued yet.  Server API should default to
        # the client API.
        payload = self.transport.payloads[-1]
        self.assertMessages(payload["messages"], [])
        self.assertEqual(payload.get("client-api"), CLIENT_API)
        self.assertEqual(payload.get("server-api"), SERVER_API)
        self.assertEqual(self.transport.message_api, SERVER_API)

        self.mstore.add({"type": "a", "api": "1.0"})
        self.mstore.add({"type": "b", "api": "1.0"})
        self.mstore.add({"type": "c", "api": "1.1"})
        self.mstore.add({"type": "d", "api": "1.1"})

        # Simulate an old 2.0 client, which has no API on messages.
        self.mstore.add({"type": "e", "api": None})
        self.mstore.add({"type": "f", "api": None})

        self.exchanger.exchange()

        payload = self.transport.payloads[-1]
        self.assertMessages(payload["messages"],
                            [{"type": "a", "api": "1.0"},
                             {"type": "b", "api": "1.0"}])
        self.assertEqual(payload.get("client-api"), CLIENT_API)
        self.assertEqual(payload.get("server-api"), "1.0")
        self.assertEqual(self.transport.message_api, "1.0")

        self.exchanger.exchange()

        payload = self.transport.payloads[-1]
        self.assertMessages(payload["messages"],
                            [{"type": "c", "api": "1.1"},
                             {"type": "d", "api": "1.1"}])
        self.assertEqual(payload.get("client-api"), CLIENT_API)
        self.assertEqual(payload.get("server-api"), "1.1")
        self.assertEqual(self.transport.message_api, "1.1")

        self.exchanger.exchange()

        payload = self.transport.payloads[-1]
        self.assertMessages(payload["messages"],
                            [{"type": "e", "api": None},
                             {"type": "f", "api": None}])
        self.assertEqual(payload.get("client-api"), CLIENT_API)
        self.assertEqual(payload.get("server-api"), "2.0")
        self.assertEqual(self.transport.message_api, "2.0")

    def test_exchange_token(self):
        """
        When sending messages to the server, the exchanger provides the
        token that the server itself gave it during the former exchange.
        """
        self.exchanger.exchange()
        self.assertIs(None, self.transport.exchange_token)
        exchange_token = self.mstore.get_exchange_token()
        self.assertIsNot(None, exchange_token)
        self.exchanger.exchange()
        self.assertEqual(exchange_token, self.transport.exchange_token)

    def test_reset_exchange_token_on_failure(self):
        """
        If an exchange fails we set the value of the next exchange token to
        C{None}, so we can authenticate ourselves even if we couldn't receive
        a valid token.
        """
        self.mstore.set_exchange_token("abcd-efgh")
        self.mstore.commit()
        self.transport.exchange = lambda *args, **kwargs: None
        self.exchanger.exchange()
        # Check that the change was persisted
        persist = Persist(filename=self.persist_filename)
        store = MessageStore(persist, self.config.message_store_path)
        self.assertIs(None, store.get_exchange_token())

    def test_include_total_messages_none(self):
        """
        The payload includes the total number of messages that the client has
        pending for the server.
        """
        self.mstore.set_accepted_types(["empty"])
        self.exchanger.exchange()
        self.assertEqual(self.transport.payloads[0]["total-messages"], 0)

    def test_include_total_messages_some(self):
        """
        If there are no more messages than those that are sent in the exchange,
        the total-messages is equivalent to the number of messages sent.
        """
        self.mstore.set_accepted_types(["empty"])
        self.mstore.add({"type": "empty"})
        self.exchanger.exchange()
        self.assertEqual(self.transport.payloads[0]["total-messages"], 1)

    def test_include_total_messages_more(self):
        """
        If there are more messages than those that are sent in the exchange,
        the total-messages is equivalent to the total number of messages
        pending.
        """
        exchanger = MessageExchange(self.reactor, self.mstore, self.transport,
                                    self.identity, self.exchange_store,
                                    self.config, max_messages=1)
        self.mstore.set_accepted_types(["empty"])
        self.mstore.add({"type": "empty"})
        self.mstore.add({"type": "empty"})
        exchanger.exchange()
        self.assertEqual(self.transport.payloads[0]["total-messages"], 2)

    def test_impending_exchange(self):
        """
        A reactor event is emitted shortly (10 seconds) before an exchange
        occurs.
        """
        self.exchanger.schedule_exchange()
        events = []
        self.reactor.call_on("impending-exchange", lambda: events.append(True))
        self.wait_for_exchange(delta=-11)
        self.assertEqual(events, [])
        self.reactor.advance(1)
        self.assertEqual(events, [True])

    def test_impending_exchange_on_urgent(self):
        """
        The C{impending-exchange} event is fired 10 seconds before urgent
        exchanges.
        """
        # We create our own MessageExchange because the one set up by the text
        # fixture has an urgent exchange interval of 10 seconds, which makes
        # testing this awkward.
        self.config.urgent_exchange_interval = 20
        exchanger = MessageExchange(self.reactor, self.mstore, self.transport,
                                    self.identity, self.exchange_store,
                                    self.config)
        exchanger.schedule_exchange(urgent=True)
        events = []
        self.reactor.call_on("impending-exchange", lambda: events.append(True))
        self.reactor.advance(9)
        self.assertEqual(events, [])
        self.reactor.advance(1)
        self.assertEqual(events, [True])

    def test_impending_exchange_gets_reschudeled_with_urgent_reschedule(self):
        """
        When an urgent exchange is scheduled after a regular exchange was
        scheduled but before it executed, the old C{impending-exchange} event
        should be cancelled and a new one should be scheduled for 10 seconds
        before the new urgent exchange.
        """
        self.config.exchange_interval = 60 * 60
        self.config.urgent_exchange_interval = 20
        exchanger = MessageExchange(self.reactor, self.mstore, self.transport,
                                    self.identity, self.exchange_store,
                                    self.config)
        events = []
        self.reactor.call_on("impending-exchange", lambda: events.append(True))
        # This call will:
        # * schedule the exchange for an hour from now
        # * schedule impending-exchange to be fired an hour - 10 seconds from
        #   now
        exchanger.schedule_exchange()
        # And this call will:
        # * hopefully cancel those previous calls
        # * schedule an exchange for 20 seconds from now
        # * schedule impending-exchange to be fired in 10 seconds
        exchanger.schedule_exchange(urgent=True)
        self.reactor.advance(10)
        self.assertEqual(events, [True])
        self.reactor.advance(10)
        self.assertEqual(len(self.transport.payloads), 1)
        # Now the urgent exchange should be fired, which should automatically
        # schedule a regular exchange.
        # Let's make sure that that *original* impending-exchange event has
        # been cancelled:
        self.reactor.advance(60 * 60  # time till exchange
                             - 10  # time till notification
                             - 20)  # time that we've already advanced
        self.assertEqual(events, [True])
        # Ok, so no new events means that the original call was
        # cancelled. great.
        # Just a bit more sanity checking:
        self.reactor.advance(20)
        self.assertEqual(events, [True, True])
        self.reactor.advance(10)
        self.assertEqual(len(self.transport.payloads), 2)

    def test_pre_exchange_event(self):
        reactor_mock = self.mocker.patch(self.reactor)
        reactor_mock.fire("pre-exchange")
        self.mocker.replay()
        self.exchanger.exchange()

    def test_schedule_exchange(self):
        self.exchanger.schedule_exchange()
        self.wait_for_exchange(urgent=True)
        self.assertFalse(self.transport.payloads)
        self.wait_for_exchange()
        self.assertTrue(self.transport.payloads)

    def test_schedule_urgent_exchange(self):
        self.exchanger.schedule_exchange(urgent=True)
        self.wait_for_exchange(urgent=True)
        self.assertTrue(self.transport.payloads)

    def test_exchange_failed_fires_correctly(self):
        """
        Ensure that the exchange-failed event is fired if the
        exchanger raises an exception.
        """

        def failed_send_data(*args, **kwargs):
            return None

        self.transport.exchange = failed_send_data

        exchanged = []

        def exchange_failed_callback():
            exchanged.append(True)

        self.reactor.call_on("exchange-failed", exchange_failed_callback)
        self.exchanger.exchange()
        self.assertEqual(exchanged, [True])

    def test_stop(self):
        self.exchanger.schedule_exchange()
        self.exchanger.stop()
        self.wait_for_exchange()
        self.assertFalse(self.transport.payloads)

    def test_stop_twice_doesnt_break(self):
        self.exchanger.schedule_exchange()
        self.exchanger.stop()
        self.exchanger.stop()
        self.wait_for_exchange()
        self.assertFalse(self.transport.payloads)

    def test_set_intervals(self):
        """
        When a C{set-intervals} message is received, the runtime attributes of
        the L{MessageExchange} are changed, the configuration values as well,
        and the configuration is written to disk to be persisted.
        """
        server_message = [{"type": "set-intervals",
                           "urgent-exchange": 1234, "exchange": 5678}]
        self.transport.responses.append(server_message)

        self.exchanger.exchange()

        self.assertEqual(self.config.exchange_interval, 5678)
        self.assertEqual(self.config.urgent_exchange_interval, 1234)

        new_config = BrokerConfiguration()
        new_config.load_configuration_file(self.config_filename)
        self.assertEqual(new_config.exchange_interval, 5678)
        self.assertEqual(new_config.urgent_exchange_interval, 1234)

    def test_set_intervals_with_urgent_exchange_only(self):
        server_message = [{"type": "set-intervals", "urgent-exchange": 1234}]
        self.transport.responses.append(server_message)

        self.exchanger.exchange()

        # Let's make sure it works.
        self.exchanger.schedule_exchange(urgent=True)
        self.reactor.advance(1233)
        self.assertEqual(len(self.transport.payloads), 1)
        self.reactor.advance(1)
        self.assertEqual(len(self.transport.payloads), 2)

    def test_set_intervals_with_exchange_only(self):
        server_message = [{"type": "set-intervals", "exchange": 5678}]
        self.transport.responses.append(server_message)

        self.exchanger.exchange()

        # Let's make sure it works.
        self.reactor.advance(5677)
        self.assertEqual(len(self.transport.payloads), 1)
        self.reactor.advance(1)
        self.assertEqual(len(self.transport.payloads), 2)

    def test_register_message(self):
        """
        The exchanger expsoses a mechanism for subscribing to messages
        of a particular type.
        """
        messages = []
        self.exchanger.register_message("type-A", messages.append)
        msg = {"type": "type-A", "whatever": 5678}
        server_message = [msg]
        self.transport.responses.append(server_message)
        self.exchanger.exchange()
        self.assertEqual(messages, [msg])

    def test_register_multiple_message_handlers(self):
        """
        Registering multiple handlers for the same type will cause
        each handler to be called in the order they were registered.
        """
        messages = []

        def handler1(message):
            messages.append(("one", message))

        def handler2(message):
            messages.append(("two", message))

        self.exchanger.register_message("type-A", handler1)
        self.exchanger.register_message("type-A", handler2)

        msg = {"type": "type-A", "whatever": 5678}
        server_message = [msg]
        self.transport.responses.append(server_message)
        self.exchanger.exchange()
        self.assertEqual(messages, [("one", msg), ("two", msg)])

    def test_server_uuid_is_stored_on_message_store(self):
        self.transport.extra["server-uuid"] = "first-uuid"
        self.exchanger.exchange()
        self.assertEqual(self.mstore.get_server_uuid(), "first-uuid")
        self.transport.extra["server-uuid"] = "second-uuid"
        self.exchanger.exchange()
        self.assertEqual(self.mstore.get_server_uuid(), "second-uuid")

    def test_server_uuid_change_cause_event(self):
        called = []

        def server_uuid_changed(old_uuid, new_uuid):
            called.append((old_uuid, new_uuid))
        self.reactor.call_on("server-uuid-changed", server_uuid_changed)

        # Set it for the first time, and it should emit the event
        # letting the system know about the change.
        self.transport.extra["server-uuid"] = "first-uuid"
        self.exchanger.exchange()
        self.assertEqual(len(called), 1)
        self.assertEqual(called[-1], (None, "first-uuid"))

        # Using the same one again, nothing should happen:
        self.transport.extra["server-uuid"] = "first-uuid"
        self.exchanger.exchange()
        self.assertEqual(len(called), 1)

        # Changing it, we should get an event again:
        self.transport.extra["server-uuid"] = "second-uuid"
        self.exchanger.exchange()
        self.assertEqual(len(called), 2)
        self.assertEqual(called[-1], ("first-uuid", "second-uuid"))

        # And then, it shouldn't emit it once more, since it continues
        # to be the same.
        self.transport.extra["server-uuid"] = "second-uuid"
        self.exchanger.exchange()
        self.assertEqual(len(called), 2)

    def test_server_uuid_event_not_emitted_with_matching_stored_uuid(self):
        """
        If the UUID in the message store is the same as the current UUID,
        the event is not emitted.
        """
        called = []

        def server_uuid_changed(old_uuid, new_uuid):
            called.append((old_uuid, new_uuid))
        self.reactor.call_on("server-uuid-changed", server_uuid_changed)

        self.mstore.set_server_uuid("the-uuid")
        self.transport.extra["server-uuid"] = "the-uuid"
        self.exchanger.exchange()
        self.assertEqual(called, [])

    def test_server_uuid_change_is_logged(self):
        self.transport.extra["server-uuid"] = "the-uuid"
        self.exchanger.exchange()

        self.assertIn("INFO: Server UUID changed (old=None, new=the-uuid).",
                      self.logfile.getvalue())

        # An exchange with the same UUID shouldn't be logged.
        self.logfile.truncate(0)
        self.transport.extra["server-uuid"] = "the-uuid"
        self.exchanger.exchange()

        self.assertNotIn("INFO: Server UUID changed", self.logfile.getvalue())

    def test_return_messages_have_their_context_stored(self):
        """
        Incoming messages with an 'operation-id' key will have the secure id
        stored in the L{ExchangeStore}.
        """
        messages = []
        self.exchanger.register_message("type-R", messages.append)
        msg = {"type": "type-R", "whatever": 5678, "operation-id": 123456}
        server_message = [msg]
        self.transport.responses.append(server_message)
        self.exchanger.exchange()
        [message] = messages
        self.assertIsNot(
            None,
            self.exchange_store.get_message_context(message['operation-id']))
        message_context = self.exchange_store.get_message_context(
            message['operation-id'])
        self.assertEqual(message_context.operation_id, 123456)
        self.assertEqual(message_context.message_type, "type-R")

    def test_one_way_messages_do_not_have_their_context_stored(self):
        """
        Incoming messages without an 'operation-id' key will *not* have the
        secure id stored in the L{ExchangeStore}.
        """
        ids_before = self.exchange_store.all_operation_ids()

        msg = {"type": "type-R", "whatever": 5678}
        server_message = [msg]
        self.transport.responses.append(server_message)
        self.exchanger.exchange()

        ids_after = self.exchange_store.all_operation_ids()
        self.assertEqual(ids_before, ids_after)

    def test_obsolete_response_messages_are_discarded(self):
        """
        An obsolete response message will be discarded as opposed to being
        sent to the server.

        A response message is considered obsolete if the secure ID changed
        since the request message was received.
        """
        # Receive the message below from the server.
        msg = {"type": "type-R", "whatever": 5678, "operation-id": 234567}
        server_message = [msg]
        self.transport.responses.append(server_message)
        self.exchanger.exchange()

        # Change the secure ID so that the response message gets discarded.
        self.identity.secure_id = 'brand-new'
        ids_before = self.exchange_store.all_operation_ids()

        self.mstore.set_accepted_types(["resynchronize"])
        message_id = self.exchanger.send(
            {"type": "resynchronize", "operation-id": 234567})
        self.exchanger.exchange()
        self.assertEqual(2, len(self.transport.payloads))
        messages = self.transport.payloads[1]["messages"]
        self.assertEqual([], messages)
        self.assertIs(None, message_id)
        expected_log_entry = (
            "Response message with operation-id 234567 was discarded because "
            "the client's secure ID has changed in the meantime")
        self.assertIn(expected_log_entry, self.logfile.getvalue())

        # The MessageContext was removed after utilisation.
        ids_after = self.exchange_store.all_operation_ids()
        self.assertEqual(len(ids_after), len(ids_before) - 1)
        self.assertNotIn('234567', ids_after)

    def test_error_exchanging_causes_failed_exchange(self):
        """
        If a traceback occurs whilst exchanging, the 'exchange-failed'
        event should be fired.
        """
        events = []

        def failed_exchange():
            events.append(None)

        self.reactor.call_on("exchange-failed", failed_exchange)
        self.exchanger._transport = RaisingTransport()
        self.exchanger.exchange()
        self.assertEqual([None], events)

    def test_error_exchanging_records_failure_in_message_store(self):
        """
        If a traceback occurs whilst exchanging, the failure is recorded
        in the message store.
        """
        mock_message_store = self.mocker.proxy(self.mstore)
        mock_message_store.record_failure(MATCH(lambda x: type(x) is int))
        self.mocker.result(None)
        self.mocker.replay()

        exchanger = MessageExchange(
            self.reactor, mock_message_store, RaisingTransport(),
            self.identity, self.exchange_store, self.config)
        exchanger.exchange()

    def test_error_exchanging_marks_exchange_complete(self):
        """
        If a traceback occurs whilst exchanging, the exchange is still
        marked as complete.
        """
        events = []

        def exchange_done():
            events.append(None)

        self.reactor.call_on("exchange-done", exchange_done)
        self.exchanger._transport = RaisingTransport()
        self.exchanger.exchange()
        self.assertEqual([None], events)

    def test_error_exchanging_logs_failure(self):
        """
        If a traceback occurs whilst exchanging, the failure is logged.
        """
        self.exchanger._transport = RaisingTransport()
        self.exchanger.exchange()
        self.assertIn("Message exchange failed.", self.logfile.getvalue())


class AcceptedTypesMessageExchangeTest(LandscapeTest):

    helpers = [ExchangeHelper]

    def setUp(self):
        super(AcceptedTypesMessageExchangeTest, self).setUp()
        self.pinger = Pinger(self.reactor, self.identity, self.exchanger,
                             self.config)
        # The __init__ method of RegistrationHandler registers a few default
        # message types that we want to catch as well
        self.handler = RegistrationHandler(
            self.config, self.identity, self.reactor, self.exchanger,
            self.pinger, self.mstore, fetch_async)

    def test_register_accepted_message_type(self):
        self.exchanger.register_client_accepted_message_type("type-B")
        self.exchanger.register_client_accepted_message_type("type-A")
        self.exchanger.register_client_accepted_message_type("type-C")
        self.exchanger.register_client_accepted_message_type("type-A")
        types = self.exchanger.get_client_accepted_message_types()
        self.assertEqual(types,
                         sorted(["type-A", "type-B", "type-C"] +
                                DEFAULT_ACCEPTED_TYPES))

    def test_exchange_sends_message_type_when_no_hash(self):
        self.exchanger.register_client_accepted_message_type("type-A")
        self.exchanger.register_client_accepted_message_type("type-B")
        self.exchanger.exchange()
        self.assertEqual(
            self.transport.payloads[0]["client-accepted-types"],
            sorted(["type-A", "type-B"] + DEFAULT_ACCEPTED_TYPES))

    def test_exchange_does_not_send_message_types_when_hash_matches(self):
        self.exchanger.register_client_accepted_message_type("type-A")
        self.exchanger.register_client_accepted_message_type("type-B")
        types = sorted(["type-A", "type-B"] + DEFAULT_ACCEPTED_TYPES)
        accepted_types_digest = md5(";".join(types)).digest()
        self.transport.extra["client-accepted-types-hash"] = \
            accepted_types_digest
        self.exchanger.exchange()
        self.exchanger.exchange()
        self.assertNotIn("client-accepted-types", self.transport.payloads[1])

    def test_exchange_continues_sending_message_types_on_no_hash(self):
        """
        If the server does not respond with a hash of client accepted message
        types, the client will continue to send the accepted types.
        """
        self.exchanger.register_client_accepted_message_type("type-A")
        self.exchanger.register_client_accepted_message_type("type-B")
        self.exchanger.exchange()
        self.exchanger.exchange()
        self.assertEqual(
            self.transport.payloads[1]["client-accepted-types"],
            sorted(["type-A", "type-B"] + DEFAULT_ACCEPTED_TYPES))

    def test_exchange_sends_new_accepted_types_hash(self):
        """
        If the accepted types on the client change between exchanges, the
        client will send a new list to the server.
        """
        self.exchanger.register_client_accepted_message_type("type-A")
        types_hash = md5("type-A").digest()
        self.transport.extra["client-accepted-types-hash"] = types_hash
        self.exchanger.exchange()
        self.exchanger.register_client_accepted_message_type("type-B")
        self.exchanger.exchange()
        self.assertEqual(
            self.transport.payloads[1]["client-accepted-types"],
            sorted(["type-A", "type-B"] + DEFAULT_ACCEPTED_TYPES))

    def test_exchange_sends_new_types_when_server_screws_up(self):
        """
        If the server suddenly and without warning changes the hash of
        accepted client types that it sends to the client, the client will
        send a new list of types.
        """
        self.exchanger.register_client_accepted_message_type("type-A")
        types_hash = md5("type-A").digest()
        self.transport.extra["client-accepted-types-hash"] = types_hash
        self.exchanger.exchange()
        self.transport.extra["client-accepted-types-hash"] = "lol"
        self.exchanger.exchange()
        self.exchanger.exchange()
        self.assertEqual(
            self.transport.payloads[2]["client-accepted-types"],
            sorted(["type-A"] + DEFAULT_ACCEPTED_TYPES))

    def test_register_message_adds_accepted_type(self):
        """
        Using the C{register_message} method of the exchanger causes
        the registered message to be included in the accepted types of
        the client that are sent to the server.
        """
        self.exchanger.register_message("typefoo", lambda m: None)
        types = self.exchanger.get_client_accepted_message_types()
        self.assertEqual(types, sorted(["typefoo"] + DEFAULT_ACCEPTED_TYPES))


class GetAcceptedTypesDiffTest(LandscapeTest):

    def test_diff_empty(self):
        self.assertEqual(get_accepted_types_diff([], []), "")

    def test_diff_add(self):
        self.assertEqual(get_accepted_types_diff([], ["wubble"]), "+wubble")

    def test_diff_remove(self):
        self.assertEqual(get_accepted_types_diff(["wubble"], []), "-wubble")

    def test_diff_no_change(self):
        self.assertEqual(get_accepted_types_diff(["ooga"], ["ooga"]), "ooga")

    def test_diff_complex(self):
        self.assertEqual(get_accepted_types_diff(["foo", "bar"],
                                                 ["foo", "ooga"]),
                         "+ooga foo -bar")



from landscape.tests.helpers import LandscapeTest
from landscape.lib.dns import (
    _lookup_server_record, _lookup_hostname, discover_server)

from twisted.internet import defer
from twisted.names import dns
from twisted.names.error import ResolverError


class FakeResolverResult(object):
    """
    A fake resolver result returned by L{FakeResolver}.

    @param type: The result type L{twisted.names.dns.SRV}
    @param payload: The result contents
    """
    def __init__(self):
        self.type = None

        class Payload(object):
            """
            A payload result returned by fake resolver.

            @param target: The result of the lookup
            """
            def __init__(self):
                self.target = ""

        class Target(object):
            """
            A payload target returned by fake resolver.

            @param name: The name contained by the target.
            """
            def __init__(self):
                self.name = ""

        self.payload = Payload()
        self.payload.target = Target()


class FakeResolver(object):
    """
    A fake resolver that mimics L{twisted.names.client.Resolver}
    """
    def __init__(self):
        self.results = None
        self.name = None
        self.queried = None

    def lookupService(self, arg1):
        self.queried = arg1
        deferred = defer.Deferred()
        deferred.callback(self.results)
        return deferred

    def getHostByName(self, arg1):
        self.queried = arg1
        deferred = defer.Deferred()
        deferred.callback(self.name)
        return deferred


class BadResolver(object):
    """
    A resolver that mimics L{twisted.names.client.Resolver} and always returns
    an error.
    """
    def lookupService(self, arg1):
        deferred = defer.Deferred()
        deferred.errback(ResolverError("Couldn't connect"))
        return deferred

    def getHostByName(self, arg1):
        deferred = defer.Deferred()
        deferred.errback(ResolverError("Couldn't connect"))
        return deferred


class DnsSrvLookupTest(LandscapeTest):
    def test_with_server_found(self):
        """
        Looking up a DNS SRV record should return the result of the lookup.
        """
        fake_result = FakeResolverResult()
        fake_result.type = dns.SRV
        fake_result.payload.target.name = "a.b.com"
        fake_resolver = FakeResolver()
        fake_resolver.results = [[fake_result]]
        query_string = "_landscape._tcp.mylandscapehost.com"

        def check(result):
            self.assertEqual(fake_resolver.queried, query_string)
            self.assertEqual("a.b.com", result)

        d = _lookup_server_record(fake_resolver, query_string)
        d.addCallback(check)
        return d

    def test_with_server_not_found(self):
        """
        Looking up a DNS SRV record and finding nothing exists should return
        an empty string.
        """
        fake_resolver = FakeResolver()
        fake_resolver.results = [[]]

        def check(result):
            self.assertEqual("", result)

        d = _lookup_server_record(fake_resolver,
                                 "_landscape._tcp.mylandscapehost.com")
        d.addCallback(check)
        return d

    def test_with_resolver_error(self):
        """A resolver error triggers error handling code."""
        # The failure should be properly logged
        logging_mock = self.mocker.replace("logging.info")
        logging_mock("SRV lookup of _landscape._tcp.mylandscapehost.com "
                     "failed.")
        self.mocker.replay()

        d = _lookup_server_record(BadResolver(),
                                 "_landscape._tcp.mylandscapehost.com")
        self.assertFailure(d, ResolverError)
        return d


class DnsNameLookupTest(LandscapeTest):
    def test_with_name_found(self):
        """
        Looking up a DNS name record should return the result of the lookup.
        """
        fake_resolver = FakeResolver()
        fake_resolver.name = "a.b.com"
        query_string = "landscape.localdomain"

        def check(result):
            self.assertEqual(fake_resolver.queried, query_string)
            self.assertEqual("a.b.com", result)

        d = _lookup_hostname(None, fake_resolver, query_string)
        d.addCallback(check)
        return d

    def test_with_name_not_found(self):
        """
        Looking up a DNS NAME record and not finding a result should return
        None.
        """
        fake_resolver = FakeResolver()
        fake_resolver.name = None

        def check(result):
            self.assertEqual(None, result)

        d = _lookup_hostname(None, fake_resolver, "landscape.localdomain")
        d.addCallback(check)
        return d

    def test_with_resolver_error(self):
        """A resolver error triggers error handling code."""
        # The failure should be properly logged
        logging_mock = self.mocker.replace("logging.info")
        logging_mock("Name lookup of landscape.localdomain failed.")
        self.mocker.replay()

        d = _lookup_hostname(None, BadResolver(), "landscape.localdomain")
        self.assertFailure(d, ResolverError)
        return d


class DiscoverServerTest(LandscapeTest):
    def test_srv_lookup(self):
        """The DNS name of the server is found using a SRV lookup."""
        fake_result = FakeResolverResult()
        fake_result.type = dns.SRV
        fake_result.payload.target.name = "a.b.com"
        fake_resolver = FakeResolver()
        fake_resolver.results = [[fake_result]]

        d = discover_server(resolver=fake_resolver)

        def check(result):
            self.assertEqual("a.b.com", result)

        d.addCallback(check)
        return d

    def test_a_name_lookup(self):
        """The DNS name of the server is found using an A name lookup."""
        fake_resolver = FakeResolver()
        fake_resolver.name = "x.y.com"

        d = discover_server(resolver=fake_resolver)

        def check(result):
            self.assertEqual("x.y.com", result)

        d.addCallback(check)
        return d

    def test_failed_lookup(self):
        """A resolver error is returned when server autodiscovery fails."""
        d = _lookup_server_record(BadResolver(), "landscape.localdomain")
        self.assertFailure(d, ResolverError)
        return d



import time

from landscape.lib.monitor import CoverageMonitor
from landscape.lib.sysstats import MemoryStats

from landscape.accumulate import Accumulator
from landscape.monitor.plugin import MonitorPlugin


class MemoryInfo(MonitorPlugin):
    """Plugin captures information about free memory and free swap."""

    persist_name = "memory-info"
    scope = "memory"
    # Prevent the Plugin base-class from scheduling looping calls.
    run_interval = None

    def __init__(self, interval=15, monitor_interval=60 * 60,
                 source_filename="/proc/meminfo", create_time=time.time):
        self._interval = interval
        self._monitor_interval = monitor_interval
        self._source_filename = source_filename
        self._memory_info = []
        self._create_time = create_time

    def register(self, registry):
        super(MemoryInfo, self).register(registry)
        self._accumulate = Accumulator(self._persist, self.registry.step_size)
        self.registry.reactor.call_every(self._interval, self.run)
        self._monitor = CoverageMonitor(self._interval, 0.8,
                                        "memory/swap snapshot",
                                        create_time=self._create_time)
        self.registry.reactor.call_every(self._monitor_interval,
                                         self._monitor.log)
        self.registry.reactor.call_on("stop", self._monitor.log, priority=2000)
        self.call_on_accepted("memory-info", self.send_message, True)

    def create_message(self):
        memory_info = self._memory_info
        self._memory_info = []
        return {"type": "memory-info", "memory-info": memory_info}

    def send_message(self, urgent=False):
        message = self.create_message()
        if len(message["memory-info"]):
            self.registry.broker.send_message(
                message, self._session_id, urgent=urgent)

    def exchange(self, urgent=False):
        self.registry.broker.call_if_accepted("memory-info",
                                              self.send_message, urgent)

    def run(self):
        self._monitor.ping()
        new_timestamp = int(self._create_time())
        memstats = MemoryStats(self._source_filename)
        memory_step_data = self._accumulate(
            new_timestamp, memstats.free_memory, "accumulate-memory")
        swap_step_data = self._accumulate(
            new_timestamp, memstats.free_swap, "accumulate-swap")

        if memory_step_data and swap_step_data:
            timestamp = memory_step_data[0]
            free_memory = int(memory_step_data[1])
            free_swap = int(swap_step_data[1])
            self._memory_info.append((timestamp, free_memory, free_swap))
